from decimal import Decimal
import bloqade.ir.location as location
import bloqade.ir.control.sequence as sequence
import bloqade.ir.control.pulse as pulse
import bloqade.ir.control.field as field
import bloqade.ir.analog_circuit as analog_circuit
import bloqade.ir.control.waveform as waveform
from bloqade.ir.location import (
    Chain,
    Honeycomb,
    Kagome,
    Lieb,
    ListOfLocations,
    Rectangular,
    Square,
    Triangular,
)
from bloqade.ir.location.base import AtomArrangement, LocationInfo, ParallelRegister
import bloqade.ir.scalar as scalar

from bloqade.ir.visitor.analog_circuit import AnalogCircuitVisitor
from bloqade.ir.visitor.location import LocationVisitor
from bloqade.ir.visitor.waveform import WaveformVisitor
from bloqade.ir.visitor.scalar import ScalarVisitor

import json

from typing import Any, Dict


class ScalarSerilaizer(ScalarVisitor):
    def visit_literal(self, ast: scalar.Literal) -> Dict[str, Dict[str, str]]:
        return {"literal": {"value": str(ast.value)}}

    def visit_variable(self, ast: scalar.Variable) -> Dict[str, Dict[str, str]]:
        return {"variable": {"name": ast.name}}

    def visit_assigned_variable(self, ast: scalar.AssignedVariable) -> Dict[str, Any]:
        return {
            "assigned_variable": {
                "name": ast.name,
                "value": str(ast.value),
            }
        }

    def visit_negative(self, ast: scalar.Negative) -> Dict[str, Any]:
        return {"negative": {"expr": self.visit(ast.expr)}}

    def visit_add(self, ast: scalar.Add) -> Dict[str, Any]:
        return {"add": {"lhs": self.visit(ast.left), "rhs": self.visit(ast.right)}}

    def visit_mul(self, ast: scalar.Mul) -> Dict[str, Any]:
        return {"mul": {"lhs": self.visit(ast.lhs), "rhs": self.visit(ast.rhs)}}

    def visit_div(self, ast: scalar.Max) -> Dict[str, Any]:
        return {"div": {"lhs": self.visit(ast.lhs), "rhs": self.visit(ast.rhs)}}

    def visit_min(self, ast: scalar.Min) -> Dict[str, Any]:
        return {"min": {"exprs": list(map(self.visit, ast.exprs))}}

    def visit_max(self, ast: scalar.Min) -> Dict[str, Any]:
        return {"max": {"exprs": list(map(self.visit, ast.exprs))}}

    def visit_slice(self, ast: scalar.Slice):
        return {
            "slice_scalar": {
                "expr": self.visit(ast.expr),
                "interval": self.visit(ast.interval),
            }
        }

    def visit_interval(self, ast: scalar.Interval) -> Dict[str, Any]:
        match (ast.start, ast.stop):
            case (None, _):
                return {"interval": {"stop": self.visit(ast.stop)}}
            case (_, None):
                return {"interval": {"start": self.visit(ast.start)}}
            case (_, _):
                return {
                    "interval": {
                        "start": self.visit(ast.start),
                        "stop": self.visit(ast.stop),
                    }
                }
            case _:
                raise ValueError(f"Invalid Interval({ast.start}, {ast.stop})")

    def default(self, obj: Any) -> Dict[str, Any]:
        if isinstance(obj, scalar.Scalar):
            return self.visit(obj)

        return super().default(obj)


class WaveformSerializer(WaveformVisitor):
    def __init__(self):
        self.scalar_encoder = ScalarSerilaizer()

    def visit_constant(self, ast: waveform.Constant) -> Dict[str, Any]:
        return {"constant": {"value": self.scalar_encoder.visit(ast.value)}}

    def visit_linear(self, ast: waveform.Linear) -> Dict[str, Any]:
        return {
            "linear": {
                "start": self.scalar_encoder.visit(ast.start),
                "stop": self.scalar_encoder.visit(ast.stop),
                "duration": self.scalar_encoder.visit(ast.duration),
            }
        }

    def visit_poly(self, ast: waveform.Poly) -> Dict[str, Any]:
        return {
            "poly": {
                "coeffs": list(map(self.scalar_encoder.visit, ast.coeffs)),
                "duration": self.scalar_encoder.visit(ast.duration),
            }
        }

    def visit_python_fn(self, ast: waveform.PythonFn) -> Dict[str, Any]:
        raise ValueError("Bloqade does not support serialization of Python code.")

    def visit_negative(self, ast: waveform.Negative) -> Dict[str, Any]:
        return {"negative": {"waveform": self.visit(ast.waveform)}}

    def visit_add(self, ast: waveform.Add) -> Dict[str, Any]:
        return {"add": {"left": self.visit(ast.left), "right": self.visit(ast.right)}}

    def visit_scale(self, ast: waveform.Scale) -> Dict[str, Any]:
        return {
            "scale": {
                "waveform": self.visit(ast.waveform),
                "scalar": self.scalar_encoder.visit(ast.scalar),
            }
        }

    def visit_slice(self, ast: waveform.Slice) -> Dict[str, Any]:
        return {
            "slice_waveform": {
                "waveform": self.visit(ast.waveform),
                "interval": self.scalar_encoder.visit(ast.interval),
            }
        }

    def visit_sample(self, ast: waveform.Sample) -> Dict[str, Any]:
        return {
            "sample": {
                "waveform": self.visit(ast.waveform),
                "dt": self.scalar_encoder.visit(ast.dt),
                "interpolation": ast.interpolation.value,
            }
        }

    def visit_append(self, ast: waveform.Append) -> Dict[str, Any]:
        return {"append_waveform": {"waveforms": list(map(self.visit, ast.waveforms))}}

    def visit_record(self, ast: waveform.Record) -> Dict[str, Any]:
        return {
            "record": {
                "var": self.scalar_encoder(ast.var),
                "waveform": self.visit(ast.waveform),
            }
        }

    def visit_smooth(self, ast: waveform.Smooth) -> Dict[str, Any]:
        match ast.kernel:
            case waveform.Gaussian:
                kernel = "gaussian"
            case waveform.Logistic:
                kernel = "logistic"
            case waveform.Sigmoid:
                kernel = "sigmoid"
            case waveform.Triangle:
                kernel = "triangle"
            case waveform.Uniform:
                kernel = "uniform"
            case waveform.Parabolic:
                kernel = "parabolic"
            case waveform.Biweight:
                kernel = "biweight"
            case waveform.Triweight:
                kernel = "triweight"
            case waveform.Tricube:
                kernel = "tricube"
            case waveform.Cosine:
                kernel = "cosine"

        return {
            "smooth": {
                "waveform": self.visit(ast.waveform),
                "radius": self.scalar_encoder.visit(ast.radius),
                "kernel": kernel,
            }
        }

    def visit_alligned(self, ast: waveform.AlignedWaveform) -> Dict[str, Any]:
        if isinstance(ast.value, scalar.Scalar):
            value = self.scalar_encoder.visit(ast.value)
        else:
            value = ast.value.value
        return {
            "alligned": {
                "waveform": self.visit(ast.waveform),
                "allignment": self.scalar_encoder.visit(ast.allignment),
                "value": value,
            }
        }

    def default(self, obj: Any) -> Dict[str, Any]:
        if isinstance(obj, waveform.Waveform):
            return self.visit(obj)

        return super().default(obj)


class LocationSerializer(LocationVisitor):
    def __init__(self):
        self.scalar_serializer = ScalarSerilaizer()

    def visit_location_info(self, info: LocationInfo) -> Any:
        return {
            "location_info": {
                "position": list(map(self.scalar_serializer.visit, info.position)),
                "filled": bool(info.filling.value),
            }
        }

    def visit_chain(self, ast: Chain) -> Any:
        return {
            "chain": {
                "lattice_spacing": self.scalar_serializer.visit(ast.lattice_spacing),
                "L": ast.shape[0],
            }
        }

    def visit_honeycomb(self, ast: Honeycomb) -> Any:
        return {
            "honeycomb": {
                "lattice_spacing": self.scalar_serializer.visit(ast.lattice_spacing),
                "L": ast.shape[0],
            }
        }

    def visit_kagome(self, ast: Kagome) -> Any:
        return {
            "kagome": {
                "lattice_spacing": self.scalar_serializer.visit(ast.lattice_spacing),
                "L": ast.shape[0],
            }
        }

    def visit_lieb(self, ast: Lieb) -> Any:
        return {
            "lieb": {
                "lattice_spacing": self.scalar_serializer.visit(ast.lattice_spacing),
                "L": ast.shape[0],
            }
        }

    def visit_list_of_locations(self, ast: ListOfLocations) -> Any:
        return {
            "list_of_locations": {
                "location_list": list(map(self.visit, ast.location_list))
            }
        }

    def visit_rectangular(self, ast: Rectangular) -> Any:
        return {
            "rectangular": {
                "lattice_spacing_x": self.scalar_serializer.visit(
                    ast.lattice_spacing_x
                ),
                "lattice_spacing_y": self.scalar_serializer.visit(
                    ast.lattice_spacing_y
                ),
                "width": ast.shape[0],
                "height": ast.shape[1],
            }
        }

    def visit_square(self, ast: Square) -> Any:
        return {
            "square": {
                "lattice_spacing": self.visit(
                    self.scalar_serializer.visit(ast.lattice_spacing)
                ),
                "L": ast.shape[0],
            }
        }

    def visit_triangular(self, ast: Triangular) -> Any:
        return {
            "triangular": {
                "lattice_spacing": self.scalar_serializer.visit(ast.lattice_spacing),
                "L": ast.shape[0],
            }
        }

    def visit_parallel_register(self, ast: location.ParallelRegister) -> Any:
        return {
            "parallel_register": {
                "register": self.visit(ast._register),
                "cluster_spacing": self.visit(ast._cluster_spacing),
            }
        }


class ProgramSerializer(AnalogCircuitVisitor):
    def __init__(self) -> None:
        self.waveform_serializer = WaveformSerializer()
        self.scalar_serializer = ScalarSerilaizer()
        self.location_serializer = LocationSerializer()

    def visit_register(self, ast: AtomArrangement) -> Any:
        return self.location_serializer.visit(ast)

    def visit_parallel_register(self, ast: ParallelRegister) -> Any:
        return self.location_serializer.visit(ast)

    def visit_sequence(self, ast: sequence.SequenceExpr) -> Any:
        match ast:
            case sequence.Sequence(pulses):
                object_map_str = {
                    sequence.rydberg: "rydberg",
                    sequence.hyperfine: "hyperfine",
                }
                return {
                    "sequence": {
                        "pulses": {
                            object_map_str[k]: self.visit(v) for k, v in pulses.items()
                        }
                    }
                }
            case sequence.Append(sequences):
                return {
                    "append_sequence": {"sequences": [self.visit(s) for s in sequences]}
                }
            case sequence.NamedSequence(sub_sequence, name):
                return {
                    "named_sequence": {
                        "name": name,
                        "sub_sequence": self.visit(sub_sequence),
                    }
                }
            case sequence.Slice(sub_sequence, interval):
                return {
                    "slice_sequence": {
                        "sequence": self.visit(sub_sequence),
                        "interval": self.scalar_serializer.visit(interval),
                    }
                }

    def visit_pulse(self, ast: pulse.PulseExpr) -> Any:
        match ast:
            case pulse.Pulse(fields):
                object_map_str = {
                    pulse.detuning: "detuning",
                    pulse.rabi.amplitude: "rabi_frequency_amplitude",
                    pulse.rabi.phase: "rabi_frequency_phase",
                }
                return {
                    "pulse": {
                        "fields": {
                            object_map_str[k]: self.visit(v) for k, v in fields.items()
                        }
                    }
                }
            case pulse.Append(pulses):
                return {"append_pulse": {"pulses": [self.visit(p) for p in pulses]}}
            case pulse.NamedPulse(name, sub_pulse):
                return {
                    "named_pulse": {"name": name, "sub_pulse": self.visit(sub_pulse)}
                }
            case pulse.Slice(sub_pulse, interval):
                return {
                    "slice_pulse": {
                        "pulse": self.visit(sub_pulse),
                        "interval": self.scalar_serializer.visit(interval),
                    }
                }

    def visit_spatial_modulation(self, ast: field.SpatialModulation) -> Any:
        match ast:
            case field.ScaledLocations(value):
                return {
                    "scaled_locations": [
                        (self.visit(k), self.visit(v)) for k, v in value.items()
                    ]
                }
            case field.Uniform:
                return "uniform"
            case field.RunTimeVector(name):
                return {"run_time_vector": {"name": name}}
            case field.AssignedRunTimeVector(name, value):
                return {
                    "assigned_run_time_vector": {
                        "name": name,
                        "value": [str(v) for v in value],
                    }
                }

    def visit_field(self, ast: field.Field) -> Any:
        return {
            "field": {
                "value": [(self.visit(k), self.visit(v)) for k, v in ast.value.items()]
            }
        }

    def visit_waveform(self, ast: waveform.Waveform) -> Any:
        return self.waveform_serializer.visit(ast)

    def visit_analog_circuit(self, ast: analog_circuit.AnalogCircuit) -> Any:
        return {
            "bloqade_program": {
                "sequence": self.visit(ast.sequence),
                "register": self.visit(ast.register),
                "static_params": ast.static_params,
                "batch_params": ast.batch_params,
                "order": ast.order,
            }
        }


class BloqadeIRSerializer(json.JSONEncoder):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.program_serializer = ProgramSerializer()
        self.waveform_serializer = WaveformSerializer()
        self.scalar_serializer = ScalarSerilaizer()
        self.bloqade_seq_types = (
            analog_circuit.AnalogCircuit,
            location.AtomArrangement,
            location.ParallelRegister,
            analog_circuit.AnalogCircuit,
            sequence.SequenceExpr,
            pulse.PulseExpr,
            field.FieldExpr,
        )

    def default(self, o: Any) -> Any:
        if isinstance(o, self.bloqade_seq_types):
            return self.program_serializer.visit(o)
        elif isinstance(o, waveform.Waveform):
            return self.waveform_serializer.visit(o)
        elif isinstance(o, scalar.Scalar):
            return self.scalar_serializer.visit(o)
        else:
            return super().default(o)


class BloqadeIRDeserializer:
    def is_register_obj(self, obj: Dict[str, Any]) -> bool:
        return (
            "list_of_locations" in obj
            or "location_info" in obj
            or "chain" in obj
            or "square" in obj
            or "honeycomb" in obj
            or "triangular" in obj
            or "lieb" in obj
            or "kagome" in obj
            or "rectangular" in obj
            or "parallel_register" in obj
        )

    def is_sequence_obj(self, obj: Dict[str, Any]) -> bool:
        return (
            "sequence" in obj
            or "append_sequence" in obj
            or "named_sequence" in obj
            or "slice_sequence" in obj
        )

    def is_pulse_obj(self, obj: Dict[str, Any]) -> bool:
        return (
            "pulse" in obj
            or "append_pulse" in obj
            or "named_pulse" in obj
            or "slice_pulse" in obj
        )

    def is_field_obj(self, obj: Dict[str, Any]) -> bool:
        return (
            "field" in obj
            or "scaled_locations" in obj
            or "uniform" in obj
            or "run_time_vector" in obj
            or "assigned_run_time_vector" in obj
        )

    def is_waveform_obj(self, obj: Dict[str, Any]) -> bool:
        return (
            "constant" in obj
            or "linear" in obj
            or "poly" in obj
            or "negative" in obj
            or "add" in obj
            or "scale" in obj
            or "slice_waveform" in obj
            or "sample" in obj
            or "append_waveform" in obj
            or "record" in obj
            or "smooth" in obj
            or "alligned" in obj
        )

    def is_scalar_obj(self, obj: Dict[str, Any]) -> bool:
        return (
            "literal" in obj
            or "variable" in obj
            or "assigned_variable" in obj
            or "negative" in obj
            or "add" in obj
            or "mul" in obj
            or "div" in obj
            or "min" in obj
            or "max" in obj
            or "slice_scalar" in obj
            or "interval" in obj
        )

    def is_bloqade_ir(self, obj: Dict[str, Any]) -> bool:
        return (
            self.is_scalar_obj(obj)
            or self.is_waveform_obj(obj)
            or self.is_field_obj(obj)
            or self.is_pulse_obj(obj)
            or self.is_sequence_obj(obj)
            or self.is_register_obj(obj)
            or "bloqade_program" in obj
        )

    def scalar_hook(self, obj: Dict[str, Any]):
        from decimal import Decimal

        match obj:
            case {"literal": {"value": str(value)}}:
                return scalar.Literal(Decimal(value))
            case {"variable": {"name": str(name)}}:
                return scalar.Variable(name)
            case {"assigned_variable": {"name": str(name), "value": str(value)}}:
                return scalar.AssignedVariable(name, Decimal(value))
            case {"negative": {"expr": expr}}:
                return scalar.Negative(expr)
            case {"add": {"lhs": lhs, "rhs": rhs}}:
                return scalar.Add(lhs, rhs)
            case {"mul": {"lhs": lhs, "rhs": rhs}}:
                return scalar.Mul(lhs, rhs)
            case {"div": {"lhs": lhs, "rhs": rhs}}:
                return scalar.Div(lhs, rhs)
            case {"min": {"exprs": exprs}}:
                return scalar.Min(frozenset(exprs))
            case {"max": {"exprs": exprs}}:
                return scalar.Max(frozenset(exprs))
            case {"slice_scalar": {"expr": expr, "interval": interval}}:
                return scalar.Slice(expr, interval)
            case {"interval": {"start": start, "stop": stop}}:
                return scalar.Interval(start, stop)
            case {"interval": {"start": start}}:
                return scalar.Interval(start, None)
            case {"interval": {"stop": stop}}:
                return scalar.Interval(None, stop)
            case _:
                return obj

    def waveform_hook(self, obj: Dict[str, Any]):
        match obj:
            case {"constant": {"value": value}}:
                return waveform.Constant(BloqadeIRDeserializer.object_hook(value))
            case {"linear": {"start": start, "stop": stop, "duration": duration}}:
                return waveform.Linear(start, stop, duration)
            case {"poly": {"coeffs": coeffs, "duration": duration}}:
                return waveform.Poly(coeffs, duration)
            case {"negative": {"waveform": sub_waveform}}:
                return waveform.Negative(sub_waveform)
            case {"add": {"left": left, "right": right}}:
                return waveform.Add(left, right)
            case {"scale": {"waveform": sub_waveform, "scalar": scale}}:
                return waveform.Scale(sub_waveform, scale)
            case {"slice_waveform": {"waveform": sub_waveform, "interval": interval}}:
                return waveform.Slice(sub_waveform, interval)
            case {
                "sample": {
                    "waveform": sub_waveform,
                    "dt": dt,
                    "interpolation": interpolation,
                }
            }:
                return waveform.Sample(sub_waveform, dt, interpolation)
            case {"append_waveform": {"waveforms": waveforms}}:
                return waveform.Append(waveforms)
            case {"record": {"var": var, "waveform": sub_waveform}}:
                return waveform.Record(var, sub_waveform)
            case {
                "smooth": {
                    "waveform": sub_waveform,
                    "radius": radius,
                    "kernel": kernel_str,
                }
            }:
                match kernel_str:
                    case "gaussian":
                        kernel = waveform.Gaussian
                    case "logistic":
                        kernel = waveform.Logistic
                    case "sigmoid":
                        kernel = waveform.Sigmoid
                    case "triangle":
                        kernel = waveform.Triangle
                    case "uniform":
                        kernel = waveform.Uniform
                    case "parabolic":
                        kernel = waveform.Parabolic
                    case "biweight":
                        kernel = waveform.Biweight
                    case "triweight":
                        kernel = waveform.Triweight
                    case "tricube":
                        kernel = waveform.Tricube
                    case "cosine":
                        kernel = waveform.Cosine
                    case _:
                        raise ValueError(f"Invalid kernel: {kernel_str}")

                return waveform.Smooth(radius, kernel, sub_waveform)
            case {
                "alligned": {
                    "waveform": sub_waveform,
                    "allignment": allignment,
                    "value": value_obj,
                }
            }:
                match value_obj:
                    case str():
                        value = waveform.AlignedValue(value_obj)
                    case _:
                        value = value_obj

                return waveform.AlignedWaveform(sub_waveform, allignment, value)
            case _:
                raise ValueError(f"Invalid waveform json: {obj}")

    def register_hook(self, obj: Dict[str, Any]):
        match obj:
            case {"location_info": {"position": position, "filled": filled}}:
                return location.LocationInfo(tuple(position), filled)
            case {"list_of_locations": {"location_list": location_list}}:
                return location.ListOfLocations(location_list)
            case {"chain": {"lattice_spacing": lattice_spacing, "L": L}}:
                return location.Chain(L, lattice_spacing)
            case {"square": {"lattice_spacing": lattice_spacing, "L": L}}:
                return location.Square(L, lattice_spacing)
            case {"honeycomb": {"lattice_spacing": lattice_spacing, "L": L}}:
                return location.Honeycomb(L, lattice_spacing)
            case {"triangular": {"lattice_spacing": lattice_spacing, "L": L}}:
                return location.Triangular(L, lattice_spacing)
            case {"lieb": {"lattice_spacing": lattice_spacing, "L": L}}:
                return location.Lieb(L, lattice_spacing)
            case {"kagome": {"lattice_spacing": lattice_spacing, "L": L}}:
                return location.Kagome(L, lattice_spacing)
            case {
                "rectangular": {
                    "lattice_spacing_x": lattice_spacing_x,
                    "lattice_spacing_y": lattice_spacing_y,
                    "width": width,
                    "height": height,
                }
            }:
                return location.Rectangular(
                    width,
                    height,
                    lattice_spacing_x,
                    lattice_spacing_y,
                )
            case {
                "parallel_register": {
                    "register": register,
                    "cluster_spacing": cluster_spacing,
                }
            }:
                return location.ParallelRegister(register, cluster_spacing)
            case _:
                raise ValueError(f"Invalid register json: {obj}")

    def field_hook(self, obj: Dict[str, Any]):
        match obj:
            case {"location": {"value": int(loc)}}:
                return field.Location(loc)
            case {"scaled_locations": {"pairs": pairs}}:
                return field.ScaledLocations(dict(pairs))
            case "uniform":
                return field.Uniform
            case {"run_time_vector": {"name": name}}:
                return field.RunTimeVector(name)
            case {"assigned_run_time_vector": {"name": name, "value": value}}:
                return field.AssignedRunTimeVector(name, list(map(Decimal, value)))
            case {"field": {"value": value}}:
                return field.Field(dict(value))
            case _:
                raise ValueError(f"Invalid field json: {obj}")

    def pulse_hook(self, obj: Dict[str, Any]):
        match obj:
            case {"pulse": {"fields": fields}}:
                str_map_object = {
                    "detuning": pulse.detuning,
                    "rabi_frequency_amplitude": pulse.rabi.amplitude,
                    "rabi_frequency_phase": pulse.rabi.phase,
                }
                return pulse.Pulse(
                    dict(map(lambda k, v: (str_map_object[k], k), fields))
                )
            case {"append_pulse": {"pulses": pulses}}:
                return pulse.Append(pulses)
            case {"named_pulse": {"name": name, "sub_pulse": sub_pulse}}:
                return pulse.NamedPulse(name, sub_pulse)
            case {"slice_pulse": {"pulse": sub_pulse, "interval": interval}}:
                return pulse.Slice(sub_pulse, interval)
            case _:
                raise ValueError(f"Invalid pulse json: {obj}")

    def sequence_hook(self, obj: Dict[str, Any]):
        match obj:
            case {"sequence": {"pulses": pulses}}:
                str_map_object = {
                    "rydberg": sequence.rydberg,
                    "hyperfine": sequence.hyperfine,
                }
                return sequence.Sequence(
                    dict(map(lambda k, v: (str_map_object[k], k), pulses))
                )
            case {"append_sequence": {"sequences": sequences}}:
                return sequence.Append(
                    list(map(BloqadeIRDeserializer.object_hook, sequences))
                )
            case {"named_sequence": {"name": name, "sub_sequence": sub_sequence}}:
                return sequence.NamedSequence(sub_sequence, name)
            case {"slice_sequence": {"sequence": sub_sequence, "interval": interval}}:
                return sequence.Slice(sub_sequence, interval)
            case _:
                raise ValueError(f"Invalid sequence json: {obj}")

    def object_hook(self, obj: Dict[str, Any]):
        if self.is_scalar_obj(obj):
            return self.scalar_hook(obj)
        elif self.is_waveform_obj(obj):
            return self.waveform_hook(obj)
        elif self.is_field_obj(obj):
            return self.field_hook(obj)
        elif self.is_pulse_obj(obj):
            return self.pulse_hook(obj)
        elif self.is_sequence_obj(obj):
            return self.sequence_hook(obj)
        elif self.is_register_obj(obj):
            return self.register_hook(obj)
        elif "bloqade_program" in obj:
            return analog_circuit.AnalogCircuit(
                register=self.register_hook(obj["bloqade_program"]["register"]),
                sequence=self.sequence_hook(obj["bloqade_program"]["sequence"]),
                static_params=obj["bloqade_program"]["static_params"],
                batch_params=obj["bloqade_program"]["batch_params"],
                order=tuple(obj["bloqade_program"]["order"]),
            )
        else:
            return obj



from bloqade.ir.analog_circuit import AnalogCircuit
from bloqade.ir.scalar import Literal
import bloqade.ir.control.waveform as waveform
from bloqade.ir.control.field import (
    Field,
    Location,
    SpatialModulation,
    ScaledLocations,
    RunTimeVector,
    AssignedRunTimeVector,
    Uniform,
)
from bloqade.ir.control.pulse import (
    PulseExpr,
    Pulse,
    NamedPulse,
    RabiFrequencyAmplitude,
    RabiFrequencyPhase,
    Detuning,
)
from bloqade.ir.control.sequence import (
    SequenceExpr,
    Sequence,
    NamedSequence,
    RydbergLevelCoupling,
    HyperfineLevelCoupling,
)
from bloqade.ir.location.base import AtomArrangement, ParallelRegister
from bloqade.ir.control.waveform import Record

from bloqade.ir.visitor.analog_circuit import AnalogCircuitVisitor
from bloqade.ir.visitor.waveform import WaveformVisitor

import bloqade.submission.ir.task_specification as task_spec
from bloqade.submission.ir.parallel import ParallelDecoder, ClusterLocationInfo
from bloqade.submission.ir.capabilities import QuEraCapabilities

from typing import Any, Dict, Tuple, List, Union, Optional
from bisect import bisect_left
import numbers
from decimal import Decimal
import numpy as np


class PiecewiseLinearCodeGen(WaveformVisitor):
    def __init__(self, assignments: Dict[str, Union[numbers.Real, List[numbers.Real]]]):
        self.assignments = assignments

    def visit_negative(
        self, ast: waveform.Negative
    ) -> Tuple[List[Decimal], List[Decimal]]:
        times, values = self.visit(ast.waveform)
        return times, [-value for value in values]

    def visit_scale(self, ast: waveform.Scale) -> Tuple[List[Decimal], List[Decimal]]:
        times, values = self.visit(ast.waveform)
        scaler = ast.scalar(**self.assignments)
        return times, [scaler * value for value in values]

    def visit_linear(self, ast: waveform.Linear) -> Tuple[List[Decimal], List[Decimal]]:
        duration = ast.duration(**self.assignments)
        start = ast.start(**self.assignments)
        stop = ast.stop(**self.assignments)

        return [Decimal(0), duration], [start, stop]

    def visit_constant(
        self, ast: waveform.Constant
    ) -> Tuple[List[Decimal], List[Decimal]]:
        duration = ast.duration(**self.assignments)
        value = ast.value(**self.assignments)

        return [Decimal(0), duration], [value, value]

    def visit_poly(self, ast: waveform.Poly) -> Tuple[List[Decimal], List[Decimal]]:
        match ast:
            case waveform.Poly(coeffs=coeff_exprs, duration=duration_expr) if len(
                coeff_exprs
            ) == 1:
                duration = duration_expr(**self.assignments)
                (value,) = [
                    coeff_expr(**self.assignments) for coeff_expr in coeff_exprs
                ]
                return [Decimal(0), duration], [value, value]

            case waveform.Poly(coeffs=coeff_exprs, duration=duration_expr) if len(
                coeff_exprs
            ) == 2:
                duration = duration_expr(**self.assignments)
                values = [coeff_expr(**self.assignments) for coeff_expr in coeff_exprs]

                start = values[0]
                stop = values[0] + values[1] * duration

                return [Decimal(0), duration], [start, stop]

            case waveform.Poly(coeffs=coeffs):
                order = len(coeffs) - 1
                raise ValueError(
                    "Failed to compile Waveform to piecewise linear,"
                    f"found Polynomial of order {order}."
                )

    def visit_slice(self, ast: waveform.Slice) -> Tuple[List[Decimal], List[Decimal]]:
        duration = ast.waveform.duration(**self.assignments)
        if ast.interval.start is None:
            start_time = Decimal(0)
        else:
            start_time = ast.interval.start(**self.assignments)

        if ast.interval.stop is None:
            stop_time = duration
        else:
            stop_time = ast.interval.stop(**self.assignments)

        if start_time < 0:
            raise ValueError((f"start time for slice {start_time} is smaller than 0."))

        if stop_time > duration:
            raise ValueError(
                (f"end time for slice {stop_time} is larger than duration {duration}.")
            )

        if stop_time < start_time:
            raise ValueError(
                (
                    f"end time for slice {stop_time} is smaller than "
                    f"starting value for slice {start_time}."
                )
            )

        if start_time == stop_time:
            return [Decimal(0.0), Decimal(0.0)], [Decimal(0.0), Decimal(0.0)]

        times, values = self.visit(ast.waveform)

        start_index = bisect_left(times, start_time)
        stop_index = bisect_left(times, stop_time)
        start_value = ast.waveform.eval_decimal(start_time, **self.assignments)
        stop_value = ast.waveform.eval_decimal(stop_time, **self.assignments)

        match (start_index, stop_index):
            case (0, int()) if stop_time == duration:
                absolute_times = times
            case (0, _):
                absolute_times = times[start_index:stop_index] + [stop_time]
                values = values[start_index:stop_index] + [stop_value]
            case (_, int()) if stop_time == duration:
                if start_time == times[start_index]:
                    absolute_times = times[start_index:]
                    values = values[start_index:]
                else:
                    absolute_times = [start_time] + times[start_index:]
                    values = [start_value] + values[start_index:]
            case (_, _):
                if start_time == times[start_index]:
                    absolute_times = times[start_index:stop_index] + [stop_time]
                    values = values[start_index:stop_index] + [stop_value]
                else:
                    absolute_times = (
                        [start_time] + times[start_index:stop_index] + [stop_time]
                    )
                    values = (
                        [start_value] + values[start_index:stop_index] + [stop_value]
                    )

        times = [time - start_time for time in absolute_times]

        return times, values

    def visit_append(self, ast: waveform.Append) -> Tuple[List[Decimal], List[Decimal]]:
        times, values = self.visit(ast.waveforms[0])

        for sub_expr in ast.waveforms[1:]:
            new_times, new_values = self.visit(sub_expr)

            # skip instructions with duration=0
            if new_times[-1] == Decimal(0):
                continue
            if values[-1] != new_values[0]:
                diff = abs(new_values[0] - values[-1])
                raise ValueError(
                    f"discontinuity with a jump of {diff} found when compiling to "
                    "piecewise linear."
                )

            shifted_times = [time + times[-1] for time in new_times[1:]]
            times.extend(shifted_times)
            values.extend(new_values[1:])

        return times, values

    def visit_sample(self, ast: waveform.Sample) -> Tuple[List[Decimal], List[Decimal]]:
        if ast.interpolation != waveform.Interpolation.Linear:
            raise ValueError(
                "Failed to compile waveform to piecewise linear, "
                f"found piecewise {ast.interpolation.value} interpolation."
            )
        return ast.samples(**self.assignments)

    def visit_record(self, ast: Record) -> Tuple[List[Decimal], List[Decimal]]:
        return self.visit(ast.waveform)


class PiecewiseConstantCodeGen(WaveformVisitor):
    def __init__(self, assignments: Dict[str, Union[numbers.Real, List[numbers.Real]]]):
        self.assignments = assignments

    def visit_negative(
        self, ast: waveform.Negative
    ) -> Tuple[List[Decimal], List[Decimal]]:
        times, values = self.visit(ast.waveform)
        return times, [-value for value in values]

    def visit_scale(self, ast: waveform.Scale) -> Tuple[List[Decimal], List[Decimal]]:
        times, values = self.visit(ast.waveform)
        scaler = ast.scalar(**self.assignments)
        return times, [scaler * value for value in values]

    def visit_linear(self, ast: waveform.Linear) -> Tuple[List[Decimal], List[Decimal]]:
        duration = ast.duration(**self.assignments)
        start = ast.start(**self.assignments)
        stop = ast.stop(**self.assignments)

        if start != stop:
            raise ValueError(
                "Failed to compile Waveform to piecewise constant, "
                "found non-constant Linear piecce."
            )

        return [0, duration], [start, stop]

    def visit_constant(
        self, ast: waveform.Constant
    ) -> Tuple[List[Decimal], List[Decimal]]:
        duration = ast.duration(**self.assignments)
        value = ast.value(**self.assignments)

        return [Decimal(0), duration], [value, value]

    def visit_poly(self, ast: waveform.Poly) -> Tuple[List[Decimal], List[Decimal]]:
        match ast:
            case waveform.Poly(coeffs=coeff_exprs, duration=duration_expr) if len(
                coeff_exprs
            ) == 1:
                duration = duration_expr(**self.assignments)
                (value,) = [
                    coeff_expr(**self.assignments) for coeff_expr in coeff_exprs
                ]
                return [Decimal(0), duration], [value, value]

            case waveform.Poly(coeffs=coeffs):
                order = len(coeffs) - 1
                raise ValueError(
                    "Failed to compile Waveform to piecewise constant,"
                    f"found Polynomial of order {order}."
                )

    def visit_slice(self, ast: waveform.Slice) -> Tuple[List[Decimal], List[Decimal]]:
        duration = ast.waveform.duration(**self.assignments)
        if ast.interval.start is None:
            start_time = Decimal(0)
        else:
            start_time = ast.interval.start(**self.assignments)

        if ast.interval.stop is None:
            stop_time = duration
        else:
            stop_time = ast.interval.stop(**self.assignments)

        if start_time < 0:
            raise ValueError((f"start time for slice {start_time} is smaller than 0."))

        if stop_time > duration:
            raise ValueError(
                (f"end time for slice {stop_time} is larger than duration {duration}.")
            )

        if stop_time < start_time:
            raise ValueError(
                (
                    f"end time for slice {stop_time} is smaller than "
                    f"starting value for slice {start_time}."
                )
            )

        if start_time == stop_time:
            return [Decimal(0.0), Decimal(0.0)], [Decimal(0.0), Decimal(0.0)]

        times, values = self.visit(ast.waveform)

        start_index = bisect_left(times, start_time)
        stop_index = bisect_left(times, stop_time)

        match (start_index, stop_index):
            case (0, int()) if stop_time == duration:
                absolute_times = times
            case (0, _):
                absolute_times = times[:stop_index] + [stop_time]
                values = values[:stop_index] + [values[stop_index - 1]]
            case (_, int()) if stop_time == duration:
                if start_time == times[start_index]:
                    absolute_times = times[start_index:]
                    values = values[start_index:]
                else:
                    absolute_times = [start_time] + times[start_index:]
                    values = [values[start_index - 1]] + values[start_index:]
            case (_, _):
                if start_time == times[start_index]:
                    absolute_times = times[start_index:stop_index] + [stop_time]
                    values = values[start_index:stop_index] + [values[stop_index - 1]]
                else:
                    absolute_times = (
                        [start_time] + times[start_index:stop_index] + [stop_time]
                    )
                    values = (
                        [values[start_index - 1]]
                        + values[start_index:stop_index]
                        + [values[stop_index - 1]]
                    )

        times = [time - start_time for time in absolute_times]

        return times, values

    def visit_append(self, ast: waveform.Append) -> Tuple[List[Decimal], List[Decimal]]:
        times, values = self.visit(ast.waveforms[0])

        for sub_expr in ast.waveforms[1:]:
            new_times, new_values = self.visit(sub_expr)

            # skip instructions with duration=0
            if new_times[-1] == Decimal(0):
                continue

            shifted_times = [time + times[-1] for time in new_times[1:]]
            times.extend(shifted_times)
            values[-1] = new_values[0]
            values.extend(new_values[1:])

        return times, values

    def visit_sample(self, ast: waveform.Sample) -> Tuple[List[Decimal], List[Decimal]]:
        if ast.interpolation != waveform.Interpolation.Constant:
            raise ValueError(
                "Failed to compile waveform to piecewise constant, "
                f"found piecewise {ast.interpolation.value} interpolation."
            )
        times, values = ast.samples(**self.assignments)

        values[-1] = values[-2]
        return times, values

    def visit_record(self, ast: Record) -> Tuple[List[Decimal], List[Decimal]]:
        return self.visit(ast.waveform)


class QuEraCodeGen(AnalogCircuitVisitor):
    def __init__(
        self,
        assignments: Dict[str, Union[numbers.Real, List[numbers.Real]]] = {},
        capabilities: Optional[QuEraCapabilities] = None,
    ):
        self.capabilities = capabilities
        self.assignments = assignments
        self.parallel_decoder = None
        self.lattice = None
        self.effective_hamiltonian = None
        self.rydberg = None
        self.field_name = None
        self.rabi_frequency_amplitude = None
        self.rabi_frequency_phase = None
        self.detuning = None
        self.lattice_site_coefficients = None

    @staticmethod
    def convert_time_to_SI_units(times: List[Decimal]):
        return [time * Decimal("1e-6") for time in times]

    @staticmethod
    def convert_energy_to_SI_units(values: List[Decimal]):
        return [value * Decimal("1e6") for value in values]

    @staticmethod
    def convert_position_to_SI_units(position: Tuple[Decimal]):
        return tuple(coordinate * Decimal("1e-6") for coordinate in position)

    def visit_spatial_modulation(self, ast: SpatialModulation):
        lattice_site_coefficients = []

        match ast:
            case ScaledLocations(locations):
                for location in locations.keys():
                    if (
                        location.value >= self.n_atoms
                    ):  # n_atoms is now the number of atoms in the parallelized
                        # lattice, but needs to be num. atoms in the original
                        raise ValueError(
                            f"Location({location.value}) is larger than the register."
                        )

                for atom_index in range(self.n_atoms):
                    scale = locations.get(Location(atom_index), Literal(0.0))
                    lattice_site_coefficients.append(
                        scale(**self.assignments)
                    )  # append scalars to lattice_site_coefficients

            case RunTimeVector(name):
                if len(self.assignments[name]) != self.n_atoms:
                    raise ValueError(
                        f"Coefficient list {name} doesn't match the size of register "
                        f"{self.n_atoms}."
                    )
                lattice_site_coefficients = list(self.assignments[name])
            case AssignedRunTimeVector(name, value):
                if len(value) != self.n_atoms:
                    raise ValueError(
                        f"Coefficient list {name} doesn't match the size of register "
                        f"{self.n_atoms}."
                    )
                lattice_site_coefficients = value

        self.lattice_site_coefficients = []
        if self.parallel_decoder:
            for cluster_site_info in self.parallel_decoder.mapping:
                self.lattice_site_coefficients.append(
                    lattice_site_coefficients[cluster_site_info.cluster_location_index]
                )
        else:
            self.lattice_site_coefficients = lattice_site_coefficients

    def visit_field(self, ast: Field):
        match (self.field_name, ast):  # Pulse: Dict of FieldName/Field
            # Can only have a global RabiFrequencyAmplitude
            case (RabiFrequencyAmplitude(), Field(terms)) if len(
                terms
            ) == 1 and Uniform in terms:
                times, values = PiecewiseLinearCodeGen(self.assignments).visit(
                    terms[Uniform]
                )

                times = QuEraCodeGen.convert_time_to_SI_units(times)
                values = QuEraCodeGen.convert_energy_to_SI_units(values)

                self.rabi_frequency_amplitude = task_spec.RabiFrequencyAmplitude(
                    global_=task_spec.GlobalField(times=times, values=values)
                )
            # Can only have global RabiFrequencyPhase
            case (RabiFrequencyPhase(), Field(terms)) if len(
                terms
            ) == 1 and Uniform in terms:  # has to be global
                times, values = PiecewiseConstantCodeGen(self.assignments).visit(
                    terms[Uniform]
                )

                times = QuEraCodeGen.convert_time_to_SI_units(times)

                self.rabi_frequency_phase = task_spec.RabiFrequencyAmplitude(
                    global_=task_spec.GlobalField(times=times, values=values)
                )

            # 3 possible detuning cases:
            # (global, no local),
            # (local, no global),
            # (local, global)

            ## global detuning, no local
            case (Detuning(), Field(terms)) if len(terms) == 1 and Uniform in terms:
                times, values = PiecewiseLinearCodeGen(self.assignments).visit(
                    terms[Uniform]
                )

                times = QuEraCodeGen.convert_time_to_SI_units(times)
                values = QuEraCodeGen.convert_energy_to_SI_units(values)

                self.detuning = task_spec.Detuning(
                    global_=task_spec.GlobalField(times=times, values=values)
                )

            ## local detuning, no global
            case (Detuning(), Field(terms)) if len(terms) == 1:
                ((spatial_modulation, waveform),) = terms.items()

                times, values = PiecewiseLinearCodeGen(self.assignments).visit(waveform)

                times = QuEraCodeGen.convert_time_to_SI_units(times)
                values = QuEraCodeGen.convert_energy_to_SI_units(values)

                self.visit(spatial_modulation)
                self.detuning = task_spec.Detuning(
                    global_=task_spec.GlobalField(
                        times=[0, times[-1]], values=[0.0, 0.0]
                    ),
                    local=task_spec.LocalField(
                        times=times,
                        values=values,
                        lattice_site_coefficients=self.lattice_site_coefficients,
                    ),
                )

            # local AND global detuning
            case (Detuning(), Field(terms)) if len(terms) == 2 and Uniform in terms:
                # will only be two keys
                for k in terms.keys():
                    if k == Uniform:
                        global_times, global_values = PiecewiseLinearCodeGen(
                            self.assignments
                        ).visit(terms[Uniform])
                    else:  # can be RunTimeVector or ScaledLocations
                        spatial_modulation = k
                        local_times, local_values = PiecewiseLinearCodeGen(
                            self.assignments
                        ).visit(terms[k])

                self.visit(spatial_modulation)  # just visit the non-uniform locations

                global_times = QuEraCodeGen.convert_time_to_SI_units(global_times)
                local_times = QuEraCodeGen.convert_time_to_SI_units(local_times)

                global_values = QuEraCodeGen.convert_energy_to_SI_units(global_values)
                local_values = QuEraCodeGen.convert_energy_to_SI_units(local_values)

                self.detuning = task_spec.Detuning(
                    local=task_spec.LocalField(
                        times=local_times,
                        values=local_values,
                        lattice_site_coefficients=self.lattice_site_coefficients,
                    ),
                    global_=task_spec.GlobalField(
                        times=global_times, values=global_values
                    ),
                )

            case _:
                raise NotImplementedError()

    def visit_pulse(self, ast: PulseExpr):
        match ast:
            case Pulse(fields):
                if RabiFrequencyAmplitude() in fields:
                    self.field_name = RabiFrequencyAmplitude()
                    self.visit(fields[self.field_name])

                if RabiFrequencyPhase() in fields:
                    self.field_name = RabiFrequencyPhase()
                    self.visit(fields[self.field_name])

                if Detuning() in fields:
                    self.field_name = Detuning()
                    self.visit(fields[self.field_name])

            case NamedPulse(pulse, _):
                self.visit(pulse)

            case _:
                raise ValueError(
                    "Failed to compile pulse expression to QuEra task, "
                    f"found: {repr(ast)}"
                )

        # fix-up any missing fields
        duration = 0.0

        if self.rabi_frequency_amplitude is not None:
            duration = max(duration, self.rabi_frequency_amplitude.global_.times[-1])

        if self.rabi_frequency_phase is not None:
            duration = max(duration, self.rabi_frequency_phase.global_.times[-1])

        if self.detuning is not None:
            duration = max(duration, self.detuning.global_.times[-1])

        if duration == 0.0:
            raise ValueError("No Fields found in pulse.")

        if self.rabi_frequency_amplitude is None:
            self.rabi_frequency_amplitude = task_spec.RabiFrequencyAmplitude(
                global_=task_spec.GlobalField(times=[0, duration], values=[0.0, 0.0])
            )

        if self.rabi_frequency_phase is None:
            self.rabi_frequency_phase = task_spec.RabiFrequencyPhase(
                global_=task_spec.GlobalField(times=[0, duration], values=[0.0, 0.0])
            )

        if self.detuning is None:
            self.detuning = task_spec.Detuning(
                global_=task_spec.GlobalField(times=[0, duration], values=[0.0, 0.0])
            )

        self.rydberg = task_spec.RydbergHamiltonian(
            rabi_frequency_amplitude=self.rabi_frequency_amplitude,
            rabi_frequency_phase=self.rabi_frequency_phase,
            detuning=self.detuning,
        )

    def visit_sequence(self, ast: SequenceExpr):
        match ast:
            case Sequence(pulses):
                if HyperfineLevelCoupling() in pulses:
                    raise ValueError("QuEra tasks does not support Hyperfine coupling.")

                self.visit(pulses.get(RydbergLevelCoupling(), Pulse({})))

            case NamedSequence(sequence, _):
                self.visit(sequence)

            case _:
                raise ValueError(
                    "Failed to compile sequence expression to QuEra task, "
                    f"found: {repr(ast)}"
                )

        self.effective_hamiltonian = task_spec.EffectiveHamiltonian(
            rydberg=self.rydberg
        )

    def visit_register(self, ast: AtomArrangement):
        sites = []
        filling = []

        for location_info in ast.enumerate():
            site = tuple(ele(**self.assignments) for ele in location_info.position)
            sites.append(QuEraCodeGen.convert_position_to_SI_units(site))
            filling.append(location_info.filling.value)

        self.n_atoms = len(sites)

        self.lattice = task_spec.Lattice(sites=sites, filling=filling)

    def visit_parallel_register(self, ast: ParallelRegister) -> Any:
        if self.capabilities is None:
            raise NotImplementedError(
                "Cannot parallelize register without device capabilities."
            )

        height_max = self.capabilities.capabilities.lattice.area.height / 1e-6
        width_max = self.capabilities.capabilities.lattice.area.width / 1e-6
        number_sites_max = (
            self.capabilities.capabilities.lattice.geometry.number_sites_max
        )

        register_filling = np.asarray(ast.register_filling)

        register_locations = np.asarray(
            [
                [s(**self.assignments) for s in location]
                for location in ast.register_locations
            ]
        )
        register_locations = register_locations - register_locations.min(axis=0)

        shift_vectors = np.asarray(
            [
                [s(**self.assignments) for s in shift_vector]
                for shift_vector in ast.shift_vectors
            ]
        )

        # build register by stack method because
        # shift_vectors might not be rectangular
        c_stack = [(0, 0)]
        visited = set([(0, 0)])
        mapping = []
        global_site_index = 0
        sites = []
        filling = []
        while c_stack:
            if len(mapping) + len(ast.register_locations) > number_sites_max:
                break

            cluster_index = c_stack.pop()

            shift = (
                shift_vectors[0] * cluster_index[0]
                + shift_vectors[1] * cluster_index[1]
            )

            new_register_locations = register_locations + shift

            # skip clusters that fall out of bounds
            if (
                np.any(new_register_locations < 0)
                or np.any(new_register_locations[:, 0] > width_max)
                or np.any(new_register_locations[:, 1] > height_max)
            ):
                continue

            new_cluster_indices = [
                (cluster_index[0] + 1, cluster_index[1]),
                (cluster_index[0], cluster_index[1] + 1),
                (cluster_index[0] - 1, cluster_index[1]),
                (cluster_index[0], cluster_index[1] - 1),
            ]

            for new_cluster_index in new_cluster_indices:
                if new_cluster_index not in visited:
                    visited.add(new_cluster_index)
                    c_stack.append(new_cluster_index)

            for cluster_location_index, (location, filled) in enumerate(
                zip(new_register_locations[:], register_filling)
            ):
                site = QuEraCodeGen.convert_position_to_SI_units(tuple(location))
                sites.append(site)
                filling.append(filled)

                mapping.append(
                    ClusterLocationInfo(
                        cluster_index=cluster_index,
                        global_location_index=global_site_index,
                        cluster_location_index=cluster_location_index,
                    )
                )

                global_site_index += 1

        self.lattice = task_spec.Lattice(sites=sites, filling=filling)
        self.n_atoms = len(ast.register_locations)
        self.parallel_decoder = ParallelDecoder(mapping=mapping)

    def visit_analog_circuit(self, ast: AnalogCircuit) -> Any:
        self.visit(ast.register)
        self.visit(ast.sequence)

    def emit(
        self, nshots: int, analog_circuit: AnalogCircuit
    ) -> Tuple[task_spec.QuEraTaskSpecification, Optional[ParallelDecoder]]:
        self.visit(analog_circuit)

        task_ir = task_spec.QuEraTaskSpecification(
            nshots=nshots,
            lattice=self.lattice,
            effective_hamiltonian=self.effective_hamiltonian,
        )

        return task_ir, self.parallel_decoder



import bloqade.ir as ir
from bloqade.ir.location import ListOfLocations, AtomArrangement
from bloqade.ir.location import Square
from bloqade import cast
import pytest


def test_listOfLocatoions():
    lattice = ListOfLocations(
        [(0, 0), (0, 1), (1, 0), (1, 1)]
    )  # enumerate method gives a generator that can be taken advantage of

    positions1 = set(map(lambda info: info.position, lattice.enumerate()))
    positions_expected = set(
        cast(
            [
                (0, 0),
                (1, 0),
                (0, 1),
                (1, 1),
            ]
        )
    )
    assert positions1 == positions_expected

    # scale:
    lattice = lattice.scale(3.2)

    positions1 = set(map(lambda info: info.position, lattice.enumerate()))
    positions_expected = set(
        cast(
            [
                (0, 0),
                (3.2, 0),
                (0, 3.2),
                (3.2, 3.2),
            ]
        )
    )
    assert positions1 == positions_expected


def test_addlocs_on_square():
    lattice = Square(2, lattice_spacing=1.0)
    lattice1 = lattice.add_positions([(9, 6), (4, 4)])

    # type(lattice)
    assert type(lattice1) == ir.location.ListOfLocations

    positions1 = set(map(lambda info: info.position, lattice1.enumerate()))
    positions_expected1 = set(cast([(0, 0), (1, 0), (0, 1), (1, 1), (9, 6), (4, 4)]))

    assert positions1 == positions_expected1


def test_addlocs_filling_options():
    lattice = Square(2, lattice_spacing=1.0)
    lattice = lattice.add_positions([(9, 6), (4, 4)], filling=[False, True])

    # type(lattice)
    assert type(lattice) == ir.location.ListOfLocations

    for info in lattice.enumerate():
        if info.position == cast((9, 6)):
            assert bool(info.filling) is False
        else:
            assert bool(info.filling) is True

    assert lattice.n_dims == 2


def test_internal_base_listofloc():
    lattice = AtomArrangement()

    with pytest.raises(NotImplementedError):
        set(map(lambda info: info.position, lattice.enumerate()))

    with pytest.raises(NotImplementedError):
        lattice.n_atoms

    with pytest.raises(NotImplementedError):
        lattice.n_dims



"""
from bloqade.builder.factory import (
    piecewise_linear,
    piecewise_constant,
    constant,
    linear,
)
from bloqade import cast


def test_ir_piecewise_linear():
    A = piecewise_linear([0.1, 3.8, 0.2], [-10, -7, "a", "b"])

    ## Append type ir node
    assert len(A.waveforms) == 3
    assert A.waveforms[0].duration == cast(0.1)
    assert A.waveforms[0].start == cast(-10)
    assert A.waveforms[0].stop == cast(-7)

    assert A.waveforms[1].duration == cast(3.8)
    assert A.waveforms[1].start == cast(-7)
    assert A.waveforms[1].stop == cast("a")

    assert A.waveforms[2].duration == cast(0.2)
    assert A.waveforms[2].start == cast("a")
    assert A.waveforms[2].stop == cast("b")


def test_ir_const():
    A = constant(value=3.415, duration=0.55)

    ## Constant type ir node:
    assert A.value == cast(3.415)
    assert A.duration == cast(0.55)


def test_ir_linear():
    A = linear(start=0.5, stop=3.2, duration=0.76)

    ## Linear type ir node:
    assert A.start == cast(0.5)
    assert A.stop == cast(3.2)
    assert A.duration == cast(0.76)


def test_ir_piecewise_constant():
    A = piecewise_constant(durations=[0.1, 3.8, 0.2], values=[-10, "a", "b"])

    assert A.waveforms[0].duration == cast(0.1)
    assert A.waveforms[0].value == cast(-10)

    assert A.waveforms[1].duration == cast(3.8)
    assert A.waveforms[1].value == cast("a")

    assert A.waveforms[2].duration == cast(0.2)
    assert A.waveforms[2].value == cast("b")
"""



from bloqade.ir.control.sequence import Sequence
from bloqade.builder.route import PragmaRoute
from bloqade.builder.base import Builder


class SequenceBuilder(PragmaRoute):
    __match_args__ = ("sequence", "__parent__")

    def __init__(self, sequence: Sequence, parent: Builder):
        self.sequence = sequence
        self.__parent__ = parent



from typing import Optional, Union, List, Callable
from numbers import Real
from functools import reduce
from bloqade.builder.base import Builder
from bloqade.builder.route import WaveformRoute
import bloqade.ir as ir

ScalarType = Union[float, str, ir.Scalar]


def assert_scalar(name, value) -> None:
    if not isinstance(value, (Real, str, ir.Scalar)):
        raise ValueError(f"{name} must be a float, str, or Scalar")


class WaveformAttachable(Builder):
    def linear(
        self, start: ScalarType, stop: ScalarType, duration: ScalarType
    ) -> "Linear":
        """
        Append/assign a linear waveform to the current location.

        Args:
            start (ScalarType Union[float, str]): The start value of the waveform
            stop (ScalarType Union[float, str]): The stop value of the waveform
            duration (ScalarType Union[float, str]): The duration of the waveform

        Examples:
            - specify a linear waveform for (spatial) uniform rydberg detuning
            from 0 to 1 in 0.5 us.

            >>> node = bloqade.start.rydberg.detuning.uniform
            >>> node = node.linear(start=0,stop=1,duration=0.5)

        Possible Next:

        - Possible Next <Location>:

            -> `.location(int)`
                :: creating new channel to address another location(s)

        - Possible Next <WaveForm:: current>:

            -> `.slice()`
                :: slice current waveform

            -> `.record(str)`
                :: record the value of waveform at current time

        - Possible Next <WaveForm:: append>:

            :: Append waveform into current channel

            -> `.linear()`

            -> `.constant()`

            -> `.ploy()`

            -> `.apply()`

            -> `.piecewise_linear()`

            -> `.piecewise_constant()`

            -> `.fn()`

        - Possible Next <LevelCoupling>:

            -> `.rydberg`
                :: Create/Switch to new rydberg level coupling channel

            -> `.hyperfine`
                :: Create/Switch to new hyperfine level coupling channel

        - Possible Next <Emit:: Linking Vars>:

            -> `.assign()`
                :: assign varialbe an actual value/number

            -> `.batch_assign()`
                :: create batch job with different
                sets of values assign to each variable.

        - Possible Next <Backend>:

            -> `.quera`
                :: specify QuEra backend

            -> `.braket`
                :: specify QuEra backend

        """
        return Linear(start, stop, duration, self)

    def constant(self, value: ScalarType, duration: ScalarType) -> "Constant":
        """
        Append/assign a constant waveform to the current location.

        Args:
            value (ScalarType Union[float, str]): The value of the waveform
            duration (ScalarType Union[float, str]): The duration of the waveform

        Examples:
            - specify a constant waveform of value 1 with duration 0.5
            for (spatial) uniform rydberg detuning

            >>> node = bloqade.start.rydberg.detuning.uniform
            >>> node = node.constant(value=1,duration=0.5)

        Possible Next:

        - Possible Next <Location>:

            -> `.location(int)`
                :: creating new channel to address another location(s)

        - Possible Next <WaveForm:: current>:

            -> `.slice()`
                :: slice current waveform

            -> `.record(str)`
                :: record the value of waveform at current time

        - Possible Next <WaveForm:: append>:

            :: Append waveform into current channel

            -> `.linear()`

            -> `.constant()`

            -> `.ploy()`

            -> `.apply()`

            -> `.piecewise_linear()`

            -> `.piecewise_constant()`

            -> `.fn()`

        - Possible Next <LevelCoupling>:

            -> `.rydberg`
                :: Create/Switch to new rydberg level coupling channel

            -> `.hyperfine`
                :: Create/Switch to new hyperfine level coupling channel

        - Possible Next <Emit:: Linking Vars>:

            -> `.assign()`
                :: assign varialbe an actual value/number

            -> `.batch_assign()`
                :: create batch job with different sets
                of values assign to each variable.

        - Possible Next <Backend>:

            -> `.quera`
                :: specify QuEra backend

            -> `.braket`
                :: specify QuEra backend

        """
        return Constant(value, duration, self)

    def poly(self, coeffs: ScalarType, duration: ScalarType) -> "Poly":
        """
        Append/assign a waveform with polynomial profile to the current location.
        with form:

            wv(t) = coeffs[0] + coeffs[1]*t + coeffs[2]*t^2 + ... + coeffs[n]*t^n

        Args:
            coeffs (ScalarType Union[float, str]): The coefficients of the polynomial
            duration (ScalarType Union[float, str]): The duration of the waveform

        Examples:
            - specify a second order polynomial with duration 0.5
            for (spatial) uniform rydberg detuning

            >>> node = bloqade.start.rydberg.detuning.uniform
            >>> node = node.poly(coeffs=[1,2,3],duration=0.5)

        Possible Next:

        - Possible Next <Location>:

            -> `.location(int)`
                :: creating new channel to address another location(s)

        - Possible Next <WaveForm:: current>:

            -> `.slice()`
                :: slice current waveform

            -> `.record(str)`
                :: record the value of waveform at current time

        - Possible Next <WaveForm:: append>:

            :: Append waveform into current channel

            -> `.linear()`

            -> `.constant()`

            -> `.ploy()`

            -> `.apply()`

            -> `.piecewise_linear()`

            -> `.piecewise_constant()`

            -> `.fn()`

        - Possible Next <LevelCoupling>:

            -> `.rydberg`
                :: Create/Switch to new rydberg level coupling channel

            -> `.hyperfine`
                :: Create/Switch to new hyperfine level coupling channel


        - Possible Next <Emit:: Linking Vars>:

            -> `.assign()`
                :: assign varialbe an actual value/number

            -> `.batch_assign()`
                :: create batch job with different sets
                of values assign to each variable.

        - Possible Next <Backend>:

            -> `.quera`
                :: specify QuEra backend

            -> `.braket`
                :: specify QuEra backend

        """
        return Poly(coeffs, duration, self)

    def apply(self, wf: ir.Waveform) -> "Apply":
        """
        Apply a pre-defined waveform to the current location.

        Args:
            wf (ir.Waveform): the waveform

        Examples:
            - apply a pre-defined waveform object to current sequence.

            >>> node = bloqade.start.rydberg.detuning.uniform
            >>> wv = bloqade.ir.Linear(0,10,0.5)
            >>> node = node.apply(wv)

        Possible Next:

        - Possible Next <Location>:

            -> `.location(int)`
                :: creating new channel to address another location(s)

        - Possible Next <WaveForm:: current>:

            -> `.slice()`
                :: slice current waveform

            -> `.record(str)`
                :: record the value of waveform at current time

        - Possible Next <WaveForm:: append>:

            :: Append waveform into current channel

            -> `.linear()`

            -> `.constant()`

            -> `.ploy()`

            -> `.apply()`

            -> `.piecewise_linear()`

            -> `.piecewise_constant()`

            -> `.fn()`

        - Possible Next <LevelCoupling>:

            -> `.rydberg`
                :: Create/Switch to new rydberg level coupling channel

            -> `.hyperfine`
                :: Create/Switch to new hyperfine level coupling channel


        - Possible Next <Emit:: Linking Vars>:

            -> `.assign()`
                :: assign varialbe an actual value/number

            -> `.batch_assign()`
                :: create batch job with different sets
                of values assign to each variable.

        - Possible Next <Backend>:

            -> `.quera`
                :: specify QuEra backend

            -> `.braket`
                :: specify QuEra backend

        """
        return Apply(wf, self)

    def piecewise_linear(
        self, durations: List[ScalarType], values: List[ScalarType]
    ) -> "PiecewiseLinear":
        """
        Append/assign a piecewise linear waveform to the current location.
        The durations should have number of elements = len(values) - 1.

        This function create a waveform by connecting `values[i], values[i+1]`
        with linear segments.

        Args:
            durations (List[ScalarType]): The durations of each linear segment
            values (List[ScalarType]): The values of each linear segment

        Examples:
            - specify a piecewise linear of [0,1,1,0] with duration [0.1,3.8,0.1]
            for (spatial) uniform rydberg detuning.

            >>> node = bloqade.start.rydberg.detuning.uniform
            >>> node = node.piecewise_linear(values=[0,1,1,0],durations=[0.1,3.8,0.1])

        Note:
            ScalarType can be either float or str.

        Possible Next:

        - Possible Next <Location>:

            -> `.location(int)`
                :: creating new channel to address another location(s)

        - Possible Next <WaveForm:: current>:

            -> `.slice()`
                :: slice current waveform

            -> `.record(str)`
                :: record the value of waveform at current time

        - Possible Next <WaveForm:: append>:

            :: Append waveform into current channel

            -> `.linear()`

            -> `.constant()`

            -> `.ploy()`

            -> `.apply()`

            -> `.piecewise_linear()`

            -> `.piecewise_constant()`

            -> `.fn()`

        - Possible Next <LevelCoupling>:

            -> `.rydberg`
                :: Create/Switch to new rydberg level coupling channel

            -> `.hyperfine`
                :: Create/Switch to new hyperfine level coupling channel


        - Possible Next <Emit:: Linking Vars>:

            -> `.assign()`
                :: assign varialbe an actual value/number

            -> `.batch_assign()`
                :: create batch job with different sets
                of values assign to each variable.

        - Possible Next <Backend>:

            -> `.quera`
                :: specify QuEra backend

            -> `.braket`
                :: specify QuEra backend

        """
        return PiecewiseLinear(durations, values, self)

    def piecewise_constant(
        self, durations: List[ScalarType], values: List[ScalarType]
    ) -> "PiecewiseConstant":
        """
        Append/assign a piecewise constant waveform to the current location.
        The durations should have number of elements = len(values).

        This function create a waveform of piecewise_constant of
        `values[i]` with duration `durations[i]`.

        Args:
            durations (List[ScalarType]): The durations of each constant segment
            values (List[ScalarType]): The values of each constant segment

        Note:
            ScalarType can be either float or str.

        Examples:
            - specify a piecewise constant of [0.5,1.5] with duration [0.1,3.8]
            for (spatial) uniform rydberg detuning.

            >>> node = bloqade.start.rydberg.detuning.uniform
            >>> node = node.piecewise_constant(values=[0.5,1.5],durations=[0.1,3.8])

        Possible Next:

        - Possible Next <Location>:

            -> `.location(int)`
                :: creating new channel to address another location(s)

        - Possible Next <WaveForm:: current>:

            -> `.slice()`
                :: slice current waveform

            -> `.record(str)`
                :: record the value of waveform at current time

        - Possible Next <WaveForm:: append>:

            :: Append waveform into current channel

            -> `.linear()`

            -> `.constant()`

            -> `.ploy()`

            -> `.apply()`

            -> `.piecewise_linear()`

            -> `.piecewise_constant()`

            -> `.fn()`

        - Possible Next <LevelCoupling>:

            -> `.rydberg`
                :: Create/Switch to new rydberg level coupling channel

            -> `.hyperfine`
                :: Create/Switch to new hyperfine level coupling channel


        - Possible Next <Emit:: Linking Vars>:

            -> `.assign()`
                :: assign varialbe an actual value/number

            -> `.batch_assign()`
                :: create batch job with different sets
                of values assign to each variable.


        - Possible Next <Backend>:

            -> `.quera`
                :: specify QuEra backend

            -> `.braket`
                :: specify QuEra backend

        """
        return PiecewiseConstant(durations, values, self)

    def fn(self, fn: Callable, duration: ScalarType) -> "Fn":
        """
        Append/assign a waveform defined by a python function to the current location.

        This function create a waveform with user-defined
        python function `fn(t)` with duration `duration`.

        Args:
            fn (Callable): The python function defining the waveform
            duration (ScalarType): The durations of each constant segment

        Note:
            - ScalarType can be either float or str.
            - The python function should take a single argument `t` and return a float.


        Examples:
            - create a cosine waveform with duration 0.5
            for (spatial) uniform rydberg detuning.

            >>> import numpy as np
            >>> def my_fn(t):
            >>>     return np.cos(2*np.pi*t)
            >>> node = bloqade.start.rydberg.detuning.uniform
            >>> node = node.fn(my_fn,duration=0.5)

        Possible Next:

        - Possible Next <Location>:

            -> `.location(int)`
                :: creating new channel to address another location(s)

        - Possible Next <WaveForm:: current>:

            -> `.slice()`
                :: slice current waveform

            -> `.record(str)`
                :: record the value of waveform at current time

            -> `.sample()`
                :: sample current callable at given time points

        - Possible Next <WaveForm:: append>:

            :: Append waveform into current channel

            -> `.linear()`

            -> `.constant()`

            -> `.ploy()`

            -> `.apply()`

            -> `.piecewise_linear()`

            -> `.piecewise_constant()`

            -> `.fn()`

        - Possible Next <LevelCoupling>:

            -> `.rydberg`
                :: Create/Switch to new rydberg level coupling channel

            -> `.hyperfine`
                :: Create/Switch to new hyperfine level coupling channel


        - Possible Next <Emit:: Linking Vars>:

            -> `.assign()`
                :: assign varialbe an actual value/number

            -> `.batch_assign()`
                :: create batch job with different sets
                of values assign to each variable.


        - Possible Next <Backend>:

            -> `.quera`
                :: specify QuEra backend

            -> `.braket`
                :: specify QuEra backend

        """
        return Fn(fn, duration, self)


# NOTE: waveform can refer previous pulse notes
#       or continue to specify pragma options
class Waveform(WaveformRoute, WaveformAttachable):
    pass


# mixin for slice and record
class Slicible:
    def slice(
        self,
        start: Optional[ScalarType] = None,
        stop: Optional[ScalarType] = None,
    ) -> "Slice":
        """
        Slice current waveform

        Possible Next:

        - Possible Next <Location>:

            -> `.location(int)`
                :: creating new channel to address
                another location(s)

        - Possible Next <WaveForm:: current>:

            -> `.record(str)`
                :: record the value of waveform at current time

        - Possible Next <WaveForm:: append>:

            :: Append waveform into current channel

            -> `.linear()`

            -> `.constant()`

            -> `.ploy()`

            -> `.apply()`

            -> `.piecewise_linear()`

            -> `.piecewise_constant()`

            -> `.fn()`

        - Possible Next <LevelCoupling>:

            -> `.rydberg`
                :: Create/Switch to new rydberg level coupling channel

            -> `.hyperfine`
                :: Create/Switch to new hyperfine level coupling channel


        - Possible Next <Emit:: Linking Vars>:

            -> `.assign()`
                :: assign varialbe an actual value/number

            -> `.batch_assign()`
                :: create batch job with different sets
                of values assign to each variable.


        - Possible Next <Backend>:

            -> `.quera`
                :: specify QuEra backend

            -> `.braket`
                :: specify QuEra backend
        """
        return Slice(start, stop, self)


class Recordable:
    def record(self, name: str) -> "Record":
        """
        Record the value of the current waveform to a variable.

        Possible Next:

        - Possible Next <Location>:

            -> `.location(int)`
                :: creating new channel to address
                another location(s)

        - Possible Next <WaveForm:: current>:

            -> `.slice()`
                :: slice the current waveform

            -> `.record(str)`
                :: record the value of waveform at current time

        - Possible Next <WaveForm:: append>:

            :: Append waveform into current channel

            -> `.linear()`

            -> `.constant()`

            -> `.ploy()`

            -> `.apply()`

            -> `.piecewise_linear()`

            -> `.piecewise_constant()`

            -> `.fn()`

        - Possible Next <LevelCoupling>:

            -> `.rydberg`
                :: Create/Switch to new rydberg level coupling channel

            -> `.hyperfine`
                :: Create/Switch to new hyperfine level coupling channel

        - Possible Next <Emit:: Linking Vars>:

            -> `.assign()`
                :: assign varialbe an actual value/number

            -> `.batch_assign()`
                :: create batch job with different sets
                of values assign to each variable.

        - Possible Next <Backend>:

            -> `.quera`
                :: specify QuEra backend

            -> `.braket`
                :: specify QuEra backend

        """
        return Record(name, self)


class WaveformPrimitive(Waveform, Slicible, Recordable):
    def __bloqade_ir__(self):
        raise NotImplementedError


class Linear(WaveformPrimitive):
    __match_args__ = ("_start", "_stop", "_duration", "__parent__")

    def __init__(
        self,
        start: ScalarType,
        stop: ScalarType,
        duration: ScalarType,
        parent: Optional[Builder] = None,
    ) -> None:
        super().__init__(parent)
        self._start = ir.cast(start)
        self._stop = ir.cast(stop)
        self._duration = ir.cast(duration)

    def __bloqade_ir__(self) -> ir.Linear:
        return ir.Linear(start=self._start, stop=self._stop, duration=self._duration)


class Constant(WaveformPrimitive):
    __match_args__ = ("_value", "_duration", "__parent__")

    def __init__(
        self, value: ScalarType, duration: ScalarType, parent: Optional[Builder] = None
    ) -> None:
        super().__init__(parent)
        self._value = ir.cast(value)
        self._duration = ir.cast(duration)

    def __bloqade_ir__(self) -> ir.Constant:
        return ir.Constant(value=self._value, duration=self._duration)


class Poly(WaveformPrimitive):
    __match_args__ = ("_coeffs", "_duration", "__parent__")

    def __init__(
        self,
        coeffs: List[ScalarType],
        duration: ScalarType,
        parent: Optional[Builder] = None,
    ) -> None:
        super().__init__(parent)
        self._coeffs = map(ir.cast, coeffs)
        self._duration = ir.cast(duration)

    def __bloqade_ir__(self):
        return ir.Poly(coeffs=self._coeffs, duration=self._duration)


class Apply(WaveformPrimitive):
    __match_args__ = ("_wf", "__parent__")

    def __init__(self, wf: ir.Waveform, parent: Optional[Builder] = None):
        super().__init__(parent)
        assert isinstance(wf, ir.Waveform), "input must be of type Waveform."
        self._wf = wf

    def __bloqade_ir__(self):
        return self._wf


class PiecewiseLinear(WaveformPrimitive):
    __match_args__ = ("_durations", "_values", "__parent__")

    def __init__(
        self,
        durations: List[ScalarType],
        values: List[ScalarType],
        parent: Optional[Builder] = None,
    ):
        assert (
            len(durations) == len(values) - 1
        ), "durations and values must be the same length"

        super().__init__(parent)
        self._durations = list(map(ir.cast, durations))
        self._values = list(map(ir.cast, values))

    def __bloqade_ir__(self):
        iter = zip(self._values[:-1], self._values[1:], self._durations)
        wfs = [ir.Linear(start=v0, stop=v1, duration=t) for v0, v1, t in iter]

        return reduce(lambda a, b: a.append(b), wfs)


class PiecewiseConstant(WaveformPrimitive):
    __match_args__ = ("_durations", "_values", "__parent__")

    def __init__(
        self,
        durations: List[ScalarType],
        values: List[ScalarType],
        parent: Optional[Builder] = None,
    ) -> None:
        assert len(durations) == len(
            values
        ), "durations and values must be the same length"
        super().__init__(parent)
        self._durations = list(map(ir.cast, durations))
        self._values = list(map(ir.cast, values))

    def __bloqade_ir__(self):
        iter = zip(self._values, self._durations)
        wfs = [ir.Constant(value=v, duration=t) for v, t in iter]
        return reduce(lambda a, b: a.append(b), wfs)


class Fn(WaveformPrimitive):
    __match_args__ = ("_fn", "_duration", "__parent__")

    def __init__(
        self,
        fn: Callable,
        duration: ScalarType,
        parent: Optional[Builder] = None,
    ) -> None:
        super().__init__(parent)
        self._fn = fn
        self._duration = ir.cast(duration)

    def sample(
        self, dt: ScalarType, interpolation: Union[ir.Interpolation, str, None] = None
    ):
        return Sample(dt, interpolation, self)

    def __bloqade_ir__(self):
        return ir.PythonFn(self._fn, self._duration)


# NOTE: no double-slice or double-record
class Slice(Waveform, Recordable):
    __match_args__ = ("_start", "_stop", "__parent__")

    def __init__(
        self,
        start: Optional[ScalarType] = None,
        stop: Optional[ScalarType] = None,
        parent: Optional[Builder] = None,
    ) -> None:
        super().__init__(parent)
        # NOTE: this should no raise for None
        self._start = ir.scalar.trycast(start)
        self._stop = ir.scalar.trycast(stop)


class Record(Waveform, Slicible):  # record should not be sliceable
    __match_args__ = ("_name", "__parent__")

    def __init__(
        self,
        name: str,
        parent: Optional[Builder] = None,
    ) -> None:
        super().__init__(parent)
        self._name = ir.var(name)


class Sample(Slicible, Recordable, WaveformRoute):
    __match_args__ = ("_dt", "_interpolation", "__parent__")

    def __init__(
        self,
        dt: ScalarType,
        interpolation: Union[ir.Interpolation, str, None],
        parent: Builder,
    ) -> None:
        super().__init__(parent)
        self._dt = ir.cast(dt)
        if interpolation is None:
            self._interpolation = None
        else:
            self._interpolation = ir.Interpolation(interpolation)



import os
import json


def load_config(qpu: str):
    real_path = os.path.realpath(__file__)
    real_path_list = os.path.split(real_path)[:-1]
    real_path = os.path.join(*real_path_list)

    match qpu:
        case "Aquila":
            with open(
                os.path.join(real_path, "config", "aquila_api_config.json"), "r"
            ) as f:
                return json.load(f)
        case "Mock":
            with open(
                os.path.join(real_path, "config", "mock_api_config.json"), "r"
            ) as f:
                return json.load(f)



from pydantic.dataclasses import dataclass
from pydantic import validator
from .tree_print import Printer
import re
from decimal import Decimal
import numbers

__all__ = [
    "var",
    "cast",
    "Scalar",
    "Interval",
    "Variable",
    "Literal",
]


@dataclass(frozen=True, repr=False)
class Scalar:
    """Base class for all scalar expressions.

    ```bnf
    <scalar> ::= <literal>
    | <variable>
    | <default>
    | <negative>
    | <add>
    | <mul>
    | <min>
    | <max>
    | <slice>
    | <inverval>

    <mul> ::= <scalar> '*' <scalar>
    <add> ::= <scalar> '+' <scalar>
    <min> ::= 'min' <scalar>+
    <max> ::= 'max' <scalar>+
    <slice> ::= <scalar expr> '[' <interval> ']'
    <interval> ::= <scalar expr> '..' <scalar expr>
    <real> ::= <literal> | <var>
    ```
    """

    def __call__(self, **kwargs) -> Decimal:
        match self:
            case Literal(value):
                return value
            case Variable(name):
                if name in kwargs:
                    return Decimal(str(kwargs[name]))
                else:
                    raise Exception(f"Unknown variable: {name}")
            case AssignedVariable(name, value):
                if name in kwargs:
                    raise ValueError(f"Variable {name} already assigned")
                else:
                    return value
            case Negative(expr):
                return -expr(**kwargs)
            case Add(lhs, rhs):
                return lhs(**kwargs) + rhs(**kwargs)
            case Mul(lhs, rhs):
                return lhs(**kwargs) * rhs(**kwargs)
            case Div(lhs, rhs):
                return lhs(**kwargs) / rhs(**kwargs)
            case Min(exprs):
                return min(map(lambda expr: expr(**kwargs), exprs))
            case Max(exprs):
                return max(map(lambda expr: expr(**kwargs), exprs))
            case Slice(expr, Interval(start, stop)):
                ret = stop - start
                ret <= expr(**kwargs)
                return ret
            case _:
                raise Exception(f"Unknown scalar expression: {self} ({type(self)})")

    def __add__(self, other: "Scalar") -> "Scalar":
        try:
            rhs = cast(other)
        except BaseException:
            return NotImplemented

        return self.add(rhs)

    def __radd__(self, other: "Scalar") -> "Scalar":
        try:
            lhs = cast(other)
        except BaseException:
            return NotImplemented

        return lhs.add(self)

    def __sub__(self, other: "Scalar") -> "Scalar":
        try:
            rhs = cast(other)
        except BaseException:
            return NotImplemented

        return self.sub(rhs)

    def __rsub__(self, other: "Scalar") -> "Scalar":
        try:
            lhs = cast(other)
        except BaseException:
            return NotImplemented

        return lhs.sub(self)

    def __mul__(self, other: "Scalar") -> "Scalar":
        try:
            rhs = cast(other)
        except BaseException:
            return NotImplemented

        return self.mul(rhs)

    def __rmul__(self, other: "Scalar") -> "Scalar":
        try:
            lhs = cast(other)
        except BaseException:
            return NotImplemented

        return lhs.mul(self)

    def __truediv__(self, other: "Scalar") -> "Scalar":
        try:
            rhs = cast(other)
        except BaseException:
            return NotImplemented

        return self.div(rhs)

    def __rtruediv__(self, other: "Scalar") -> "Scalar":
        try:
            lhs = cast(other)
        except BaseException:
            return NotImplemented

        return lhs.div(self)

    def __neg__(self) -> "Scalar":
        return Scalar.canonicalize(Negative(self))

    def add(self, other) -> "Scalar":
        expr = Add(lhs=self, rhs=cast(other))
        return Scalar.canonicalize(expr)

    def sub(self, other) -> "Scalar":
        expr = Add(lhs=self, rhs=-cast(other))
        return Scalar.canonicalize(expr)

    def mul(self, other) -> "Scalar":
        expr = Mul(lhs=self, rhs=cast(other))
        return Scalar.canonicalize(expr)

    def div(self, other) -> "Scalar":
        expr = Div(lhs=self, rhs=cast(other))
        return Scalar.canonicalize(expr)

    def min(self, other) -> "Scalar":
        expr = Min(exprs=frozenset({self, cast(other)}))
        return Scalar.canonicalize(expr)

    def max(self, other) -> "Scalar":
        expr = Max(exprs=frozenset({self, cast(other)}))
        return Scalar.canonicalize(expr)

    def __repr__(self) -> str:
        ph = Printer()
        ph.print(self)
        return ph.get_value()

    def _repr_pretty_(self, p, cycle):
        Printer(p).print(self, cycle)

    @staticmethod
    def canonicalize(expr: "Scalar") -> "Scalar":
        def minmax(op, exprs):
            new_exprs = set()
            for expr in exprs:
                match expr:
                    case op(exprs):
                        exprs = map(Scalar.canonicalize, exprs)
                        new_exprs.update(exprs)
                    case _:
                        expr = Scalar.canonicalize(expr)
                        new_exprs.add(expr)

            if len(new_exprs) > 1:
                return op(exprs=frozenset(new_exprs))
            else:
                (new_expr,) = new_exprs
                return new_expr

        match expr:
            case Negative(Negative(sub_expr)):
                return Scalar.canonicalize(sub_expr)
            case Negative(expr=Literal(value)) if value < 0:
                return Literal(-value)
            case Add(lhs=Literal(lhs), rhs=Literal(rhs)):
                return Literal(lhs + rhs)
            case Add(lhs=Literal(lhs), rhs=Negative(Literal(rhs))):
                return Literal(lhs - rhs)
            case Add(lhs=Negative(Literal(lhs)), rhs=Literal(rhs)):
                return Literal(rhs - lhs)
            case Add(lhs=Literal(0.0), rhs=sub_expr):
                return Scalar.canonicalize(sub_expr)
            case Add(lhs=sub_expr, rhs=Literal(0.0)):
                return Scalar.canonicalize(sub_expr)
            case Add(lhs=sub_expr, rhs=Negative(other_expr)) if sub_expr == other_expr:
                return Literal(0.0)
            case Add(lhs=Negative(sub_expr), rhs=other_expr) if sub_expr == other_expr:
                return Literal(0.0)
            case Mul(lhs=Literal(lhs), rhs=Literal(rhs)):
                return Literal(lhs * rhs)
            case Mul(lhs=Literal(0.0), rhs=_):
                return Literal(0.0)
            case Mul(lhs=_, rhs=Literal(0.0)):
                return Literal(0.0)
            case Mul(lhs=Literal(1.0), rhs=sub_expr):
                return Scalar.canonicalize(sub_expr)
            case Mul(lhs=sub_expr, rhs=Literal(1.0)):
                return Scalar.canonicalize(sub_expr)
            case Div(lhs=Literal(lhs), rhs=Literal(rhs)):
                return Literal(lhs / rhs)
            case Div(lhs=sub_expr, rhs=Literal(1.0)):
                return Scalar.canonicalize(sub_expr)
            case Min(exprs):
                return minmax(Min, exprs)
            case Max(exprs):
                return minmax(Max, exprs)
            case _:
                return expr


def check_variable_name(name: str) -> None:
    regex = "^[A-Za-z_][A-Za-z0-9_]*"
    re_match = re.match(regex, name)
    if re_match.group() != name:
        raise ValueError(f"string '{name}' is not a valid python identifier")


def cast(py) -> "Scalar":
    """
    1. cast Real number (or list/tuple of Real numbers)
    to [`Scalar Literal`][bloqade.ir.scalar.Literal].

    2. cast str (or list/tuple of Real numbers)
    to [`Scalar Variable`][bloqade.ir.scalar.Variable].

    Args:
        py (Union[str,Real,Tuple[Real],List[Real]]): python object to cast

    Returns:
        Scalar
    """
    ret = trycast(py)
    if ret is None:
        raise TypeError(f"Cannot cast {type(py)} to Scalar Literal")

    return ret


# TODO: RL: remove support on List and Tuple just use map?
#       this is making type inference much harder to parse
#       in human brain
# [KHW] it need to be there. For recursive replace for nested
#       list/tuple
def trycast(py) -> "Scalar | None":
    match py:
        case int(x) | float(x) | bool(x):
            return Literal(Decimal(str(x)))
        case Decimal():
            return Literal(py)
        case str(x):
            check_variable_name(x)
            return Variable(x)
        case list() as xs:
            return list(map(cast, xs))
        case tuple() as xs:
            return tuple(map(cast, xs))
        case Scalar():
            return py
        case numbers.Real():
            return Literal(Decimal(str(py)))

        case _:
            return


def var(py: str) -> "Variable":
    """cast string (or list/tuple of strings)
    to [`Variable`][bloqade.ir.scalar.Variable].

    Args:
        py (Union[str, List[str]]): a string or list/tuple of strings

    Returns:
       Union[Variable]
    """
    ret = tryvar(py)
    if ret is None:
        raise TypeError(f"Cannot cast {type(py)} to Variable")

    return ret


def tryvar(py) -> "Variable | None":
    match py:
        case str(x):
            check_variable_name(x)
            return Variable(x)
        # case list() as xs:
        #     return list(map(var, xs))
        # case tuple() as xs:
        #     return tuple(map(var, xs))
        case Variable():
            return py
        case _:
            return


class Real(Scalar):
    # """Base class for all real expressions."""
    pass


@dataclass(frozen=True, repr=False)
class Literal(Real):
    value: Decimal
    """Scalar Literal, which stores a decimaal value instance.

    Args:
        value (Decimal): decimal value instance

    """

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self):
        return f"{str(self.value)}"

    def children(self):
        return []

    def print_node(self):
        return f"Literal: {self.value}"

    def _repr_pretty_(self, p, cycle):
        Printer(p).print(self, cycle)


@dataclass(frozen=True, repr=False)
class Variable(Real):
    """Variable, which stores a variable name.

    Args:
        name (str): variable instance.

    """

    name: str

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self):
        return f"var({str(self.name)})"

    def children(self):
        return []

    def print_node(self):
        return f"Variable: {self.name}"

    def _repr_pretty_(self, p, cycle):
        Printer(p).print(self, cycle)

    @validator("name")
    def name_validator(cls, v):
        # removing reserved toekn check for now.
        # match v:
        #     case "config_file":
        #         raise ValueError(
        #             f'"{v}" is a reserved token, cannot create variable '
        #               'with that name'
        #         )
        #     case "clock_s":
        #         raise ValueError(
        #             f'"{v}" is a reserved token, cannot create variable '
        #               'with that name'
        #         )
        return v


@dataclass(frozen=True, repr=False)
class AssignedVariable(Scalar):
    name: str
    value: Decimal

    def __str__(self):
        return f"{self.name}"

    def children(self):
        return []

    def print_node(self):
        return f"AssignedVariable: {self.name} = {self.value}"

    @validator("name")
    def name_validator(cls, v):
        match v:
            case "config_file":
                raise ValueError(
                    f'"{v}" is a reserved token, cannot create variable with that name'
                )
            case "clock_s":
                raise ValueError(
                    f'"{v}" is a reserved token, cannot create variable with that name'
                )

        return v


@dataclass(frozen=True, repr=False)
class Negative(Scalar):
    expr: Scalar

    def __str__(self):
        return f"-({str(self.expr)})"

    def children(self):
        return [self.expr]

    def print_node(self):
        return "Negative"


@dataclass(frozen=True, repr=False)
class Interval:
    start: Scalar | None
    stop: Scalar | None

    @staticmethod
    def from_slice(s: slice) -> "Interval":
        match s:
            case slice(start=None, stop=None, step=None):
                raise ValueError("Slice must have at least one argument")
            case slice(start=None, stop=None, step=_):
                raise ValueError("Slice step must be None")
            case slice(start=None, stop=stop, step=None):
                return Interval(None, cast(stop))
            case slice(start=None, stop=stop, step=_):
                raise ValueError("Slice step must be None")
            case slice(start=start, stop=None, step=None):
                return Interval(cast(start), None)
            case slice(start=start, stop=None, step=_):
                raise ValueError("Slice step must be None")
            case slice(start=start, stop=stop, step=None):
                return Interval(cast(start), cast(stop))
            case slice(start=start, stop=stop, step=_):
                raise ValueError("Slice step must be None")

    def __repr__(self) -> str:
        ph = Printer()
        ph.print(self)
        return ph.get_value()

    def _repr_pretty_(self, p, cycle):
        Printer(p).print(self, cycle)

    def __str__(self):
        match (self.start, self.stop):
            case (None, None):
                raise ValueError("Interval must have at least one bound")
            case (None, stop):
                return f":{str(stop)}"
            case (start, None):
                return f"{str(start)}:"
            case (start, stop):
                return f"{str(self.start)}:{str(self.stop)}"

    def print_node(self):
        return "Interval"

    def children(self):
        match (self.start, self.stop):
            case (None, None):
                raise ValueError("Interval must have at least one bound")
            case (None, stop):
                return {"stop": stop}
            case (start, None):
                return {"start": start}
            case (start, stop):
                return {"start": start, "stop": stop}


@dataclass(frozen=True, repr=False)
class Slice(Scalar):
    expr: Scalar  # duration
    interval: Interval

    def __str__(self):
        return f"{str(self.expr)}[{str(self.interval)}]"

    def children(self):
        return {"Scalar": self.expr, None: self.interval}

    def print_node(self):
        return "Slice"


@dataclass(frozen=True, repr=False)
class Add(Scalar):
    lhs: Scalar
    rhs: Scalar

    def __str__(self):
        return f"({str(self.lhs)} + {str(self.rhs)})"

    def children(self):
        return [self.lhs, self.rhs]

    def print_node(self):
        return "+"


@dataclass(frozen=True, repr=False)
class Mul(Scalar):
    lhs: Scalar
    rhs: Scalar

    def __str__(self):
        return f"({str(self.lhs)} * {str(self.rhs)})"

    def children(self):
        return [self.lhs, self.rhs]

    def print_node(self):
        return "*"


@dataclass(frozen=True, repr=False)
class Div(Scalar):
    lhs: Scalar
    rhs: Scalar

    def __str__(self):
        return f"({str(self.lhs)} / {str(self.rhs)})"

    def children(self):
        return [self.lhs, self.rhs]

    def print_node(self):
        return "/"


@dataclass(frozen=True, repr=False)
class Min(Scalar):
    exprs: frozenset[Scalar]

    def children(self):
        return list(self.exprs)

    def print_node(self):
        return "min"

    def __str__(self):
        return f"scalar.Min({str(self.exprs)})"


@dataclass(frozen=True, repr=False)
class Max(Scalar):
    exprs: frozenset[Scalar]

    def children(self):
        return list(self.exprs)

    def print_node(self):
        return "max"

    def __str__(self):
        return f"scalar.Max({str(self.exprs)})"



from typing import Optional, Any
from bloqade.builder.base import Builder
from bloqade.builder.backend import BackendRoute
from bloqade.builder.parse.trait import Parse
from bloqade.ir import cast


class ParallelizeBase(Builder):
    __match_args__ = ("_cluster_spacing", "__parent__")

    def __init__(self, cluster_spacing: Any, parent: Optional[Builder] = None) -> None:
        super().__init__(parent)
        self._cluster_spacing = cast(cluster_spacing)


# the idea here is that parallelize in different parts of the builder
# will lead to different backends depending on the order of the calls


# If parallize before calling flatten restrict the API to only use the
# SubmitBackendRoute
class Parallelize(ParallelizeBase, BackendRoute, Parse):
    pass


# else use this after flatten restrict the API to only use the FlattenedBackendRoute
class ParallelizeFlatten(ParallelizeBase, BackendRoute, Parse):
    pass



from typing import Union, Optional, TYPE_CHECKING
from bloqade.builder.waveform import WaveformAttachable
from bloqade.builder.base import Builder
from bloqade.ir import Scalar
from numbers import Real

if TYPE_CHECKING:
    from bloqade.ir.control.field import UniformModulation

ScalarType = Union[float, str, Scalar]


class SpatialModulation(WaveformAttachable):
    pass


class Uniform(SpatialModulation):
    """
    The node specify a uniform spacial modulation. Which is ready to apply waveform
    (See [`Waveform`][bloqade.builder.waveform] for available waveform options)

    Examples:

        - To hit this node from the start node:

        >>> reg = bloqade.start.add_positions([(0,0),(1,1),(2,2),(3,3)])
        >>> loc = reg.rydberg.detuning.uniform

        - Apply Linear waveform:

        >>> wv = bloqade.ir.Linear(start=0,stop=1,duration=0.5)
        >>> reg = bloqade.start.add_positions([(0,0),(1,1),(2,2),(3,3)])
        >>> loc = reg.rydberg.detuning.uniform.apply(wv)

    """

    def __bloqade_ir__(self) -> "UniformModulation":
        from bloqade.ir import Uniform

        return Uniform


class Location(SpatialModulation):
    __match_args__ = ("_label", "__parent__")

    def __init__(self, label: int, parent: Optional[Builder] = None) -> None:
        assert isinstance(label, int) and label >= 0
        super().__init__(parent)
        self._label = label

    def location(self, label: int) -> "Location":
        """
        Append another location to the current location(s)

        Args:
            label (int): The label of the location

        Examples:

            - Append location 1 to the current location 0.

            >>> reg = bloqade.start.add_positions([(0,0),(1,1),(2,2),(3,3)])
            >>> loc = reg.rydberg.detuning.location(0)
            >>> loc = loc.location(1)

            - One can keep appending by concatenating location()

            >>> reg = bloqade.start.add_positions([(0,0),(1,1),(2,2),(3,3)])
            >>> loc = reg.rydberg.detuning.location(0)
            >>> loc = loc.location(1).location(2)

        - Possible Next <Location>:

            -> `...location(int).location(int)`
                :: keep adding location into current list

            -> `...location(int).scale(float)`
                :: specify scaling factor to current location
                for the preceeding waveform

        - Possible Next <WaveForm>:

            -> `...location(int).linear()`
                :: apply linear waveform

            -> `...location(int).constant()`
                :: apply constant waveform

            -> `...location(int).ploy()`
                :: apply polynomial waveform

            -> `...location(int).apply()`
                :: apply pre-constructed waveform

            -> `...location(int).piecewise_linear()`
                :: apply piecewise linear waveform

            -> `...location(int).piecewise_constant()`
                :: apply piecewise constant waveform

            -> `...location(int).fn()`
                :: apply callable as waveform.


        """
        return Location(label, self)

    def scale(self, value) -> "Scale":
        """
        Scale the preceeding waveform by the specified factor.

        Args:
            scale (float): The factor to scale (amplitude of)
            the preceeding waveform.

        Examples:

            - Scale the preceeding waveform that addressing location(0) by 1.2.

            >>> reg = bloqade.start.add_positions([(0,0),(1,1),(2,2),(3,3)])
            >>> scaled = reg.rydberg.detuning.location(0).scale(1.2)

            - Scale multiple locations with different factors.
            (ex. loc 0 by 1.2, loc 1 by 0.5)

            >>> reg = bloqade.start.add_positions([(0,0),(1,1),(2,2),(3,3)])
            >>> loc = reg.rydberg.detuning.location(0)
            >>> loc = loc.scale(1.2).location(1).scale(0.5)

            - Scale multiple locations with the same factor. (ex. loc 0 and 1 by 1.2)

            >>> reg = bloqade.start.add_positions([(0,0),(1,1),(2,2),(3,3)])
            >>> scaled = reg.rydberg.detuning.location(0).location(1).scale(1.2)


        - Possible Next <Location>:

            -> `...scale(float).location(int)`
                :: keep adding location into current list

        - Possible Next <WaveForm>:

            -> `...scale(float).linear()`
                :: apply linear waveform

            -> `...scale(float).constant()`
                :: apply constant waveform

            -> `...scale(float).ploy()`
                :: apply polynomial waveform

            -> `...scale(float).apply()`
                :: apply pre-constructed waveform(s)

            -> `...scale(float).piecewise_linear()`
                :: apply piecewise linear waveform

            -> `...scale(float).piecewise_constant()`
                :: apply piecewise constant waveform

            -> `...scale(float).fn()`
                :: apply callable as waveform.



        """
        return Scale(value, self)


# NOTE: not a spatial modulation itself because it only modifies the
#       location of the previous spatial modulation
class Scale(WaveformAttachable):
    __match_args__ = ("_value", "__parent__")

    def __init__(self, value: ScalarType, parent: Optional[Builder] = None) -> None:
        assert isinstance(value, (Real, str, Scalar))
        super().__init__(parent)
        self._value = value

    def location(self, label: int) -> "Location":
        """
        - Append another location to the current location after scale the previous one

        Args:
            label (int): The label of the location

        Examples:

            - Append location 1 after scale location 0 by 1.2.

            >>> reg = bloqade.start.add_positions([(0,0),(1,1),(2,2),(3,3)])
            >>> loc = reg.rydberg.detuning.location(0).scale(1.2)
            >>> loc = loc.location(1)

        - Possible Next <Location>:

            -> `...location(int).location(int)`
                :: keep adding location into current list

            -> `...location(int).scale(float)`
                :: specify scaling factor to current location
                for the preceeding waveform

        - Possible Next <WaveForm>:

            -> `...location(int).linear()`
                :: apply linear waveform

            -> `...location(int).constant()`
                :: apply constant waveform

            -> `...location(int).ploy()`
                :: apply polynomial waveform

            -> `...location(int).apply()`
                :: apply pre-constructed waveform

            -> `...location(int).piecewise_linear()`
                :: apply piecewise linear waveform

            -> `...location(int).piecewise_constant()`
                :: apply piecewise constant waveform

            -> `...location(int).fn()`
                :: apply callable as waveform.
        """
        return Location(label, self)


class Var(SpatialModulation):
    __match_args__ = ("_name", "__parent__")

    def __init__(self, name: str, parent: Optional[Builder] = None) -> None:
        assert isinstance(name, str)
        super().__init__(parent)
        self._name = name



# from bloqade.ir.routine.params import Params
from bloqade.ir.analog_circuit import AnalogCircuit
from bloqade.ir import AtomArrangement, ParallelRegister, Sequence

from bloqade.builder.base import Builder
from bloqade.ir.routine.params import Params


from dataclasses import dataclass
from typing import TYPE_CHECKING, Union, Tuple

if TYPE_CHECKING:
    from bloqade.ir.routine.braket import BraketServiceOptions
    from bloqade.ir.routine.quera import QuEraServiceOptions


class RoutineParse:
    def parse_register(self: "RoutineBase") -> Union[AtomArrangement, ParallelRegister]:
        return self.source.parse_register()

    def parse_sequence(self: "RoutineBase") -> Sequence:
        return self.source.parse_sequence()

    def parse_circuit(self: "RoutineBase") -> AnalogCircuit:
        return self.source.parse_circuit()

    def parse_source(self: "RoutineBase") -> Tuple[AnalogCircuit, Params]:
        return self.source.parse_source()


@dataclass(frozen=True)
class RoutineBase(RoutineParse):
    source: Builder


@dataclass(frozen=True)
class Routine(RoutineBase):
    """Result of parsing a completed Builder string."""

    @property
    def braket(self) -> "BraketServiceOptions":
        from .braket import BraketServiceOptions

        return BraketServiceOptions(source=self.source)

    @property
    def quera(self) -> "QuEraServiceOptions":
        from .quera import QuEraServiceOptions

        return QuEraServiceOptions(source=self.source)



from typing import Tuple, Union, TYPE_CHECKING
import json
import bloqade.ir as ir
from bloqade.visualization.builder_visualize import display_builder


if TYPE_CHECKING:
    from bloqade.ir import AtomArrangement, ParallelRegister, Sequence
    from bloqade.ir.analog_circuit import AnalogCircuit
    from bloqade.ir.routine.params import Params
    from bloqade.builder.base import Builder


class CompileJSON:
    def json(self: "Builder", **json_options) -> str:
        """transform the program to a JSON string."""
        from bloqade.builder.parse.json import BuilderSerializer

        return json.dumps(self, cls=BuilderSerializer, **json_options)

    # def __repr__(self):
    #     raise NotImplementedError


class ParseRegister:
    def parse_register(self: "Builder") -> Union["AtomArrangement", "ParallelRegister"]:
        """Parse the arrangement of atoms of the program."""
        from bloqade.builder.parse.builder import Parser

        return Parser().parse_register(self)


class ParseSequence:
    def parse_sequence(self: "Builder") -> "Sequence":
        """Parse the pulse sequence part of the program."""
        from bloqade.builder.parse.builder import Parser

        return Parser().parse_sequence(self)


class ParseCircuit:
    def parse_circuit(self: "Builder") -> "AnalogCircuit":
        """Parse the analog circuit from the program."""
        from bloqade.builder.parse.builder import Parser

        return Parser().parse_circuit(self)


class ParseRoutine:
    """Parse the program to return an AnalogCircuit as well as the parameters
    for the circuit.


    """

    def parse_source(self: "Builder") -> Tuple["AnalogCircuit", "Params"]:
        from bloqade.builder.parse.builder import Parser

        return Parser().parse_source(self)


class Parse(ParseRegister, ParseSequence, ParseCircuit, ParseRoutine):
    @property
    def n_atoms(self: "Builder"):
        """Return the number of atoms in the program."""
        from .builder import Parser

        register = Parser().parse_register(self)

        if isinstance(register, ir.location.ParallelRegister):
            return register._register.n_atoms
        else:
            return register.n_atoms

    def __repr__(self: "Builder"):
        from .builder import Parser

        analog_circ, metas = Parser().parse_source(self)

        return repr(analog_circ) + "\n" + repr(metas)

    def show(self, batch_id: int = 0):
        from .builder import Parser

        analog_circ, metas = Parser().parse_source(self)

        display_builder(analog_circ, metas, batch_id)



from bloqade.builder.base import Builder


class BloqadeService(Builder):
    @property
    def bloqade(self):
        """
        - Specify bloqade emulator
        - Possible Next:

            -> `...bloqade.python`
                :: Bloqade python backend

            -> `...bloqade.julia`
                :: Bloqade julia backend
        """
        return BloqadeDeviceRoute(self)


class BloqadeDeviceRoute(Builder):
    def python(self, solver: str):
        raise NotImplementedError

    def julia(self, solver: str, nthreads: int = 1):
        raise NotImplementedError



from typing import Any, Dict, TextIO, Type, Union
from bloqade.builder.base import Builder
from bloqade.codegen.common.json import BloqadeIRSerializer, BloqadeIRDeserializer
from bloqade.builder.start import ProgramStart
from bloqade.builder.sequence_builder import SequenceBuilder

# from bloqade.builder.base import Builder
from bloqade.builder.backend.bloqade import BloqadeDeviceRoute, BloqadeService
from bloqade.builder.backend.braket import BraketDeviceRoute, BraketService
from bloqade.builder.backend.quera import QuEraDeviceRoute, QuEraService
from bloqade.builder.spatial import Location, Scale, Var, Uniform
from bloqade.builder.waveform import (
    Linear,
    Constant,
    Poly,
    Fn,
    Apply,
    Slice,
    Record,
    Sample,
    PiecewiseLinear,
    PiecewiseConstant,
)
from bloqade.builder.field import Detuning, Rabi, RabiAmplitude, RabiPhase
from bloqade.builder.coupling import Rydberg, Hyperfine
from bloqade.builder.parallelize import Parallelize, ParallelizeFlatten
from bloqade.builder.assign import Assign, BatchAssign
from bloqade.builder.flatten import Flatten


class BuilderSerializer(BloqadeIRSerializer):
    def default(self, obj):
        def camel_to_snake(name: str) -> str:
            return "".join("_" + c.lower() if c.isupper() else c for c in name).lstrip(
                "_"
            )

        if isinstance(obj, ProgramStart) and type(obj) is ProgramStart:
            return "program_start"

        match obj:
            case BraketDeviceRoute(parent) | QuEraDeviceRoute(
                parent
            ) | BloqadeDeviceRoute(parent) | BraketService(parent) | QuEraService(
                parent
            ) | BloqadeService(
                parent
            ):
                return {camel_to_snake(obj.__class__.__name__): {"parent": parent}}
            case Parallelize(cluster_spacing, parent):
                return {
                    "parallelize": {"cluster_spacing": cluster_spacing},
                    "parent": parent,
                }
            case ParallelizeFlatten(cluster_spacing, parent):
                return {
                    "parallelize_flatten": {"cluster_spacing": cluster_spacing},
                    "parent": parent,
                }
            case SequenceBuilder(sequence, parent):
                return {
                    "sequence_builder": {
                        "sequence": self.default(sequence),
                        "parent": self.default(parent),
                    }
                }
            case Assign(assignments, parent) | BatchAssign(assignments, parent):
                return {
                    camel_to_snake(obj.__class__.__name__): {
                        "assignments": assignments,
                        "parent": parent,
                    }
                }
            case Flatten(order, parent):
                return {"flatten": {"order": order, "parent": parent}}
            case Constant(value, duration, parent):
                return {
                    "constant": {
                        "value": value,
                        "duration": duration,
                        "parent": parent,
                    }
                }
            case Linear(start, stop, duration, parent):
                return {
                    "linear": {
                        "start": start,
                        "stop": stop,
                        "duration": duration,
                        "parent": parent,
                    }
                }
            case Poly(coeffs, duration, parent):
                return {
                    "poly": {
                        "coeffs": coeffs,
                        "duration": duration,
                        "parent": parent,
                    }
                }
            case Fn():
                raise ValueError(
                    "Bloqade does not support serialization of Python code."
                )
            case Apply(wf, parent):
                return {"apply": {"wf": wf, "parent": parent}}
            case Slice(start, stop, parent):
                return {
                    "slice": {
                        "start": start,
                        "stop": stop,
                        "parent": parent,
                    }
                }
            case Record(name, parent):
                return {
                    "record": {
                        "name": name,
                        "parent": parent,
                    }
                }
            case Sample(dt, interpolation, parent):
                return {
                    "sample": {
                        "dt": dt,
                        "interpolation": interpolation,
                        "parent": parent,
                    }
                }
            case PiecewiseLinear(durations, values, parent):
                return {
                    "piecewise_linear": {
                        "durations": durations,
                        "values": values,
                        "parent": parent,
                    }
                }
            case PiecewiseConstant(durations, values, parent):
                return {
                    "piecewise_constant": {
                        "durations": durations,
                        "values": values,
                        "parent": parent,
                    }
                }
            case Location(label, parent):
                return {
                    "location": {
                        "label": label,
                        "parent": parent,
                    }
                }
            case Scale(value, parent):
                return {
                    "scale": {
                        "value": value,
                        "parent": parent,
                    }
                }
            case Var(name, parent):
                return {"Var": {"name": name, "parent": parent}}
            case Uniform(parent):
                return {"uniform": {"parent": parent}}
            case Detuning(parent) | RabiAmplitude(parent) | RabiPhase(parent) | Rabi(
                parent
            ) | Hyperfine(parent) | Rydberg(parent) | BraketDeviceRoute(
                parent
            ) | QuEraDeviceRoute(
                parent
            ) | BloqadeDeviceRoute(
                parent
            ) | BraketService(
                parent
            ) | QuEraService(
                parent
            ) | BloqadeService(
                parent
            ):
                return {camel_to_snake(obj.__class__.__name__): {"parent": parent}}
            case _:
                return super().default(obj)


class BuilderDeserializer(BloqadeIRDeserializer):
    methods: Dict[str, Type] = {
        "rydberg": Rydberg,
        "hyperfine": Hyperfine,
        "detuning": Detuning,
        "rabi": Rabi,
        "rabi_amplitude": RabiAmplitude,
        "rabi_phase": RabiPhase,
        "var": Var,
        "scale": Scale,
        "location": Location,
        "uniform": Uniform,
        "piecewise_constant": PiecewiseConstant,
        "piecewise_linear": PiecewiseLinear,
        "sample": Sample,
        "record": Record,
        "slice": Slice,
        "apply": Apply,
        "poly": Poly,
        "linear": Linear,
        "constant": Constant,
        "flatten": Flatten,
        "parallelize": Parallelize,
        "parallelize_flatten": ParallelizeFlatten,
        "sequence_builder": SequenceBuilder,
        "assign": Assign,
        "batch_assign": BatchAssign,
        "bloqade_device_route": BloqadeDeviceRoute,
        "bloqade_service": BloqadeService,
        "braket_device_route": BraketDeviceRoute,
        "braket_service": BraketService,
        "quera_device_route": QuEraDeviceRoute,
        "quera_service": QuEraService,
    }

    def object_hook(self, obj: Dict[str, Any]):
        match obj:
            case str("program_start"):
                return ProgramStart(None)
            case dict() if len(obj) == 1:
                ((head, options),) = obj.items()
                if head in self.methods:
                    return self.methods[head](**options)

        return super().object_hook(obj)


def load_program(filename_or_io: Union[str, TextIO]) -> Builder:
    import json

    if isinstance(filename_or_io, str):
        with open(filename_or_io, "r") as f:
            return json.load(f, object_hook=BuilderDeserializer().object_hook)
    else:
        return json.load(filename_or_io, object_hook=BuilderDeserializer().object_hook)


def save_program(filename_or_io: Union[str, TextIO], program: Builder) -> None:
    import json

    if isinstance(filename_or_io, str):
        with open(filename_or_io, "w") as f:
            json.dump(program, f, cls=BuilderSerializer)
    else:
        json.dump(program, filename_or_io, cls=BuilderSerializer)



from collections import OrderedDict
from dataclasses import dataclass
import json
from numbers import Real

from bloqade.ir.routine.base import RoutineBase
from bloqade.submission.quera import QuEraBackend
from bloqade.submission.mock import MockBackend
from bloqade.submission.quera_api_client.load_config import load_config
from bloqade.task.batch import RemoteBatch
from bloqade.task.quera import QuEraTask

from typing import Tuple, Union


@dataclass(frozen=True)
class QuEraServiceOptions(RoutineBase):
    def device(self, config_file: str | None, **api_config):
        if config_file is not None:
            with open(config_file, "r") as f:
                api_config = {**json.load(f), **api_config}

        backend = QuEraBackend(**api_config)

        return QuEraHardwareRoutine(source=self.source, backend=backend)

    def aquila(self) -> "QuEraHardwareRoutine":
        backend = QuEraBackend(**load_config("Aquila"))
        return QuEraHardwareRoutine(source=self.source, backend=backend)

    def cloud_mock(self):
        backend = QuEraBackend(**load_config("Mock"))
        return QuEraHardwareRoutine(source=self.source, backend=backend)

    def mock(self, state_file: str = ".mock_state.txt"):
        backend = MockBackend(state_file=state_file)
        return QuEraHardwareRoutine(source=self.source, backend=backend)


@dataclass(frozen=True)
class QuEraHardwareRoutine(RoutineBase):
    backend: Union[QuEraBackend, MockBackend]

    def compile(
        self,
        shots: int,
        args: Tuple[Real, ...] = (),
        name: str | None = None,
    ) -> RemoteBatch:
        """
        Compile to a RemoteBatch, which contain
            QuEra backend specific tasks.

        Args:
            shots (int): number of shots
            args (Tuple): additional arguments
            name (str): custom name of the batch

        Return:
            RemoteBatch

        """
        from bloqade.codegen.common.assign_variables import AssignAnalogCircuit

        from bloqade.codegen.common.assignment_scan import AssignmentScan
        from bloqade.codegen.hardware.quera import QuEraCodeGen

        circuit, params = self.parse_source()
        capabilities = self.backend.get_capabilities()
        circuit = AssignAnalogCircuit(params.static_params).visit(circuit)

        tasks = OrderedDict()

        for task_number, batch_params in enumerate(params.batch_assignments(*args)):
            record_params = AssignmentScan(batch_params).emit(circuit)
            final_circuit = AssignAnalogCircuit(record_params).visit(circuit)
            task_ir, parallel_decoder = QuEraCodeGen(capabilities=capabilities).emit(
                shots, final_circuit
            )

            task_ir = task_ir.discretize(capabilities)
            tasks[task_number] = QuEraTask(
                None, self.backend, task_ir, batch_params, parallel_decoder
            )

        batch = RemoteBatch(source=self.source, tasks=tasks, name=name)

        return batch

    def submit(
        self,
        shots: int,
        args: Tuple[Real, ...] = (),
        name: str | None = None,
        shuffle: bool = False,
        **kwargs,
    ) -> RemoteBatch:
        """
        Compile to a RemoteBatch, which contain
            QuEra backend specific tasks,
            and submit through QuEra service.

        Args:
            shots (int): number of shots
            args (Tuple): additional arguments
            name (str): custom name of the batch
            shuffle (bool): shuffle the order of jobs

        Return:
            RemoteBatch

        """
        batch = self.compile(shots, args, name)
        batch._submit(shuffle, **kwargs)
        return batch



from ..scalar import Scalar, cast
from ..tree_print import Printer
from .waveform import Waveform
from bokeh.plotting import figure, show
from bokeh.layouts import gridplot, row, layout
from bokeh.models.widgets import PreText
from bokeh.models import ColumnDataSource
from bloqade.visualization.ir_visualize import get_field_figure
from pydantic.dataclasses import dataclass
from typing import Dict, List
from decimal import Decimal


class FieldExpr:
    pass


__all__ = [
    "Field",
    "Location",
    "SpatialModulation",
    "Uniform",
    "RunTimeVector",
    "ScaledLocations",
]


@dataclass(frozen=True)
class Location(FieldExpr):
    value: int

    def __repr__(self) -> str:
        ph = Printer()
        ph.print(self)
        return ph.get_value()

    def _repr_pretty_(self, p, cycle):
        Printer(p).print(self, cycle)

    def __str__(self):
        return f"Location({str(self.value)})"

    def children(self):
        return []

    def print_node(self):
        return f"Location({str(self.value)})"


@dataclass
class SpatialModulation(FieldExpr):
    def __hash__(self) -> int:
        raise NotImplementedError

    def __repr__(self) -> str:
        ph = Printer()
        ph.print(self)
        return ph.get_value()

    def _repr_pretty_(self, p, cycle):
        Printer(p).print(self, cycle)

    def _get_data(self, **assignment):
        return {}

    def figure(self, **assignment):
        raise NotImplementedError


@dataclass
class UniformModulation(SpatialModulation):
    def __hash__(self) -> int:
        return hash(self.__class__)

    def __str__(self):
        return "Uniform"

    def print_node(self):
        return "UniformModulation"

    def children(self):
        return []

    def _get_data(self, **assignment):
        return ["uni"], ["all"]

    def figure(self, **assignment):
        p = figure(sizing_mode="stretch_both")
        p.text(
            x=[0.5],
            y=[0.5],
            text="Uniform",
            text_algin="center",
            text_baseline="middle",
        )
        return p

    def show(self, **assignment):
        show(self.figure(**assignment))


Uniform = UniformModulation()


@dataclass
class RunTimeVector(SpatialModulation):
    name: str

    def __hash__(self) -> int:
        return hash(self.name) ^ hash(self.__class__)

    def __str__(self):
        return f"RunTimeVector({str(self.name)})"

    def print_node(self):
        return "RunTimeVector"

    def children(self):
        return [self.name]

    def figure(self, **assginment):
        p = figure(sizing_mode="stretch_both")
        p.text(
            x=[0.5],
            y=[0.5],
            text=self.name,
            text_algin="center",
            text_baseline="middle",
        )
        return p

    def _get_data(self, **assignment):
        return [self.name], ["vec"]

    def show(self, **assignment):
        show(self.figure(**assignment))


@dataclass
class AssignedRunTimeVector(SpatialModulation):
    name: str
    value: List[Decimal]

    def __hash__(self) -> int:
        return hash(self.name) ^ hash(self.__class__) ^ hash(tuple(self.value))

    def __str__(self):
        return f"AssignedRunTimeVector({str(self.name), str(self.value)})"

    def print_node(self):
        return "AssgiendRunTimeVector"

    def children(self):
        return [self.name, *self.value]

    def figure(self, **assginment):
        p = figure(sizing_mode="stretch_both")
        p.text(
            x=[0.5],
            y=[0.5],
            text=self.name,
            text_algin="center",
            text_baseline="middle",
        )
        return p

    def _get_data(self, **assignment):
        return [self.name], ["vec"]

    def show(self, **assignment):
        show(self.figure(**assignment))


@dataclass(init=False, repr=False)
class ScaledLocations(SpatialModulation):
    value: Dict[Location, Scalar]

    def __init__(self, pairs):
        value = dict()
        for k, v in pairs.items():
            if isinstance(k, int):
                k = Location(k)
            elif isinstance(k, Location):
                pass
            else:
                raise ValueError(f"expected Location or int, got {k}")

            value[k] = cast(v)
        self.value = value

    def __hash__(self) -> int:
        return hash(frozenset(self.value.items())) ^ hash(self.__class__)

    def __str__(self):
        ## formatting location: scale pair:
        tmp = {f"{str(key.value)}": val for key, val in self.value.items()}
        return f"ScaledLocations({str(tmp)})"

    def _get_data(self, **assignments):
        names = []
        scls = []

        for loc, scl in self.value.items():
            names.append("loc[%d]" % (loc.value))
            scls.append(str(scl(**assignments)))

        return names, scls

    def print_node(self):
        return self.__str__()

    def children(self):
        # can return list or dict
        # should return dict consisting of Location and Scalar
        annotated_children = {}
        for loc, scalar in self.value.items():
            annotated_children[loc.print_node()] = scalar

        return annotated_children

    def __getitem__(self, key):
        return self.value[key]

    def __setitem__(self, key, value):
        assert isinstance(key, Location)
        self.value[key] = cast(value)

    def __bool__(self):
        return bool(self.value)

    def figure(self, **assignments):
        locs = []
        literal_val = []
        for k, v in self.value.items():
            locs.append(f"loc[{k.value}]")
            literal_val.append(float(v(**assignments)))

        source = ColumnDataSource(data=dict(locations=locs, yvals=literal_val))

        p = figure(
            y_range=locs, sizing_mode="stretch_both", x_axis_label="Scale factor"
        )
        p.hbar(y="locations", right="yvals", source=source, height=0.4)

        return p

    def show(self, **assignment):
        show(self.figure(**assignment))


@dataclass
class Field(FieldExpr):
    """Field node in the IR. Which contains collection(s) of
    [`Waveform`][bloqade.ir.control.waveform.Waveform]

    ```bnf
    <field> ::= ('field' <spatial modulation>  <padded waveform>)*
    ```
    """

    value: Dict[SpatialModulation, Waveform]

    def __hash__(self) -> int:
        return hash(frozenset(self.value.items())) ^ hash(self.__class__)

    def __repr__(self) -> str:
        ph = Printer()
        ph.print(self)
        return ph.get_value()

    def _repr_pretty_(self, p, cycle):
        Printer(p).print(self, cycle)

    def __str__(self):
        ## formatting location: scale pair:
        tmp = {str(key): str(val) for key, val in self.value.items()}

        return f"Field({str(tmp)})"

    def add(self, other):
        if not isinstance(other, Field):
            raise ValueError(f"Cannot add Field and {other.__class__}")

        out = Field(dict(self.value))

        for spatial_modulation, waveform in other.value.items():
            if spatial_modulation in self.value:
                out.value[spatial_modulation] = out.value[spatial_modulation] + waveform
            else:
                out.value[spatial_modulation] = waveform

        return out

    def print_node(self):
        return "Field"

    def children(self):
        # return dict with annotations
        return {spatial_mod.print_node(): wf for spatial_mod, wf in self.value.items()}

    def figure_old(self, **assignments):
        full_figs = []
        idx = 0
        for spmod, wf in self.value.items():
            fig_mod = spmod.figure(**assignments)
            fig_wvfm = wf.figure(**assignments)

            # format AST tree:
            txt = wf.__repr__()
            txt = "> Waveform AST:\n" + txt

            txt_asgn = ""
            # format assignment:
            if len(assignments):
                txt_asgn = "> Assignments:\n"
                for key, val in assignments.items():
                    txt_asgn += f"{key} := {val}\n"
                txt_asgn += "\n"

            # Display AST tree:

            header = "Ch[%d]\n" % (idx)
            text_box = PreText(text=header + txt_asgn + txt, sizing_mode="stretch_both")
            text_box.styles = {"overflow": "scroll", "border": "1px solid black"}

            # layout channel:
            fp = gridplot(
                [[row(text_box, fig_mod, sizing_mode="stretch_both"), fig_wvfm]],
                merge_tools=False,
                sizing_mode="stretch_both",
            )
            fp.styles = {"border": "2px solid black"}
            fp.width_policy = "max"

            full_figs.append(fp)
            idx += 1

        full = layout(
            full_figs,
            # merge_tools=False,
            sizing_mode="stretch_both",
        )
        full.width_policy = "max"

        return full

    def figure(self, **assignments):
        return get_field_figure(self, "Field", None, **assignments)

    def show(self, **assignments):
        """
        Interactive visualization of the Field

        Args:
            **assignments: assigning the instance value (literal) to the
                existing variables in the Field

        """
        show(self.figure(**assignments))



from bloqade.ir.control.waveform import Waveform, Linear, Constant
from typing import Any, List

# this part only for manually build sequences.


def linear(duration, start, stop):
    return Linear(start, stop, duration)


def constant(duration, value):
    return Constant(value, duration)


def piecewise_linear(durations: List[Any], values: List[Any]) -> Waveform:
    pwl_wf = None
    for duration, start, stop in zip(durations, values[:-1], values[1:]):
        if pwl_wf is None:
            pwl_wf = Linear(start, stop, duration)
        else:
            pwl_wf = pwl_wf.append(Linear(start, stop, duration))

    return pwl_wf


def piecewise_constant(durations: List[Any], values: List[Any]) -> Waveform:
    pwc_wf = None
    for duration, value in zip(durations, values):
        if pwc_wf is None:
            pwc_wf = Constant(value, duration)
        else:
            pwc_wf = pwc_wf.append(Constant(value, duration))

    return pwc_wf



from bokeh.models import Div
from bokeh.layouts import row, column
from bokeh.io import show


def liner(txt):
    return f"<p>{txt}</p>"


def display_builder(analog_circ, metas, batch_id: int):
    kwargs = metas.static_params
    kwargs.update(metas.batch_params[batch_id])

    out = "<p>Assignments: </p>"
    for key, val in kwargs.items():
        out += liner(f" :: {key} = {val}")

    rowobj = analog_circ.figure(**kwargs)

    field = rowobj.children[0]
    reg = rowobj.children[1]
    div = Div(text=out, width=200, height=200)

    show(row(field, column(reg, div)))



from bloqade.builder.base import Builder


class Field(Builder):
    @property
    def uniform(self):
        """
        - Addressing all atom locations for preceeding waveform
        - Next-step: <WaveForm>
        - Possible Next:

            -> `...uniform.linear()`
                :: apply linear waveform

            -> `...uniform.constant()`
                :: apply constant waveform

            -> `...uniform.ploy()`
                :: apply polynomial waveform

            -> `...uniform.apply()`
                :: apply pre-constructed waveform

            -> `...uniform.piecewise_linear()`
                :: apply piecewise linear waveform

            -> `...uniform.piecewise_constant()`
                :: apply piecewise constant waveform

            -> `...uniform.fn()`
                :: apply callable as waveform.


        Examples:

            - Addressing rydberg detuning to all atoms in the system with
            4 sites

            >>> reg = bloqade.start.add_positions([(0,0),(1,1),(2,2),(3,3)])
            >>> loc = reg.rydberg.detuning.uniform

        """
        from bloqade.builder.spatial import Uniform

        return Uniform(self)

    def location(self, label: int):
        """
        Addressing one or multiple specific location(s) for preceeding waveform.

        (See [`Location`][bloqade.builder.location.Location] for more details])

        Args:
            label (int): The label of the location to apply the following waveform to.

        Examples:

            - Addressing rydberg detuning to location 1 on a system with 4 sites.

            >>> reg = bloqade.start.add_positions([(0,0),(1,1),(2,2),(3,3)])
            >>> loc = reg.rydberg.detuning.location(1)

            - Addressing rydberg detuning on both location
            0 and 2 on a system with 4 sites.

            >>> reg = bloqade.start.add_positions([(0,0),(1,1),(2,2),(3,3)])
            >>> loc = reg.rydberg.detuning.location(1).location(2)

        Note:
            label index start with 0, and should be positive.

        - Possible Next <Location>:

            -> `...location(int).location(int)`
                :: adding location into current list

            -> `...location(int).scale(float)`
                :: specify scaling factor to current location
                for the preceeding waveform

        - Possible Next <WaveForm>:

            -> `...location(int).linear()`
                :: apply linear waveform

            -> `...location(int).constant()`
                :: apply constant waveform

            -> `...location(int).ploy()`
                :: apply polynomial waveform

            -> `...location(int).apply()`
                :: apply pre-constructed waveform

            -> `...location(int).piecewise_linear()`
                :: apply piecewise linear waveform

            -> `...location(int).piecewise_constant()`
                :: apply piecewise constant waveform

            -> `...location(int).fn()`
                :: apply callable as waveform.


        """
        from bloqade.builder.spatial import Location

        return Location(label, self)

    def var(self, name: str):
        """
        - Addressing atom location associate with given variable for preceeding waveform
        - Possible Next <WaveForm>:

            -> `...location(int).linear()`
                :: apply linear waveform

            -> `...location(int).constant()`
                :: apply constant waveform

            -> `...location(int).ploy()`
                :: apply polynomial waveform

            -> `...location(int).apply()`
                :: apply pre-constructed waveform

            -> `...location(int).piecewise_linear()`
                :: apply piecewise linear waveform

            -> `...location(int).piecewise_constant()`
                :: apply piecewise constant waveform

            -> `...location(int).fn()`
                :: apply callable as waveform.


        Examples:

            - Addressing rydberg detuning to atom location `myatom` in the system with
            4 sites

            >>> reg = bloqade.start.add_positions([(0,0),(1,1),(2,2),(3,3)])
            >>> loc = reg.rydberg.detuning.var('myatom')

        """
        from bloqade.builder.spatial import Var

        return Var(name, self)


class Detuning(Field):
    """
    This node represent detuning field of a
    specified level coupling (rydberg or hyperfine) type.


    Examples:

        - To specify detuning of rydberg coupling:

        >>> node = bloqade.start.rydberg.detuning
        >>> type(node)
        <class 'bloqade.builder.field.Detuning'>

        - To specify detuning of hyperfine coupling:

        >>> node = bloqade.start.hyperfine.detuning
        >>> type(node)
        <class 'bloqade.builder.field.Detuning'>

    Note:
        This node is a SpatialModulation node.
        See [`SpatialModulation`][bloqade.builder.field.SpatialModulation]
        for additional options.

    """

    def __bloqade_ir__(self):
        from bloqade.ir.control.pulse import detuning

        return detuning


# this is just an eye candy, thus
# it's not the actual Field object
# one can skip this node when doing
# compilation
class Rabi(Builder):
    """
    This node represent rabi field of a
    specified level coupling (rydberg or hyperfine) type.

    Examples:

        - To specify rabi of rydberg coupling:

        >>> node = bloqade.start.rydberg.rabi
        <class 'bloqade.builder.field.Rabi'>

        - To specify rabi of hyperfine coupling:

        >>> node = bloqade.start.hyperfine.rabi
        >>> type(node)
        <class 'bloqade.builder.field.Rabi'>


    """

    @property
    def amplitude(self) -> "RabiAmplitude":
        """
        - Specify the amplitude of the rabi field.
        - Next-step: <SpacialModulation>
        - Possible Next:

            -> `...amplitude.location(int)`
                :: Address atom at specific location

            -> `...amplitude.uniform`
                :: Address all atoms in register

            -> `...amplitude.var(str)`
                :: Address atom at location labeled by variable


        Examples:

            - rydberg coupling rabi amplitude
            (See also [`RabiAmplitude`][bloqade.builder.field.RabiAmplitude])

            >>> ryd_rabi = bloqade.start.rydberg.rabi
            >>> ryd_rabi_amp = ryd_rabi.amplitude


            - hyperfine coupling rabi amplitude
            (See also [`RabiAmplitude`][bloqade.builder.field.RabiAmplitude])

            >>> hyp_rabi = bloqade.start.hyperfine.rabi
            >>> hyp_rabi_amp = hyp_rabi.amplitude

        """
        return RabiAmplitude(self)

    @property
    def phase(self) -> "RabiPhase":
        """
        - Specify the phase of the rabi field.
        - Next-step: <SpacialModulation>
        - Possible Next:

            -> `...phase.location(int)`
                :: Address atom at specific location

            -> `...phase.uniform`
                :: Address all atoms in register

            -> `...phase.var(str)`
                :: Address atom at location labeled by variable


        Examples:

            - rydberg coupling rabi phase
            (See also [`RabiPhase`][bloqade.builder.field.RabiPhase])

            >>> ryd_rabi = bloqade.start.rydberg.rabi
            >>> ryd_rabi_ph = ryd_rabi.phase


            - hyperfine coupling rabi phase
            (See also [`RabiPhase`][bloqade.builder.field.RabiPhase])

            >>> hyp_rabi = bloqade.start.hyperfine.rabi
            >>> hyp_rabi_ph = hyp_rabi.phase

        """
        return RabiPhase(self)


class RabiAmplitude(Field):
    """
    This node represent amplitude of a rabi field.

    Examples:

        - To specify rabi amplitude of rydberg coupling:

        >>> node = bloqade.start.rydberg.rabi.amplitude
        >>> type(node)
        <class 'bloqade.builder.field.Amplitude'>

        - To specify rabi amplitude of hyperfine coupling:

        >>> node = bloqade.start.hyperfine.rabi.amplitude
        >>> type(node)
        <class 'bloqade.builder.field.Amplitude'>

    Note:
        This node is a SpatialModulation node.
        See [`SpatialModulation`][bloqade.builder.field.SpatialModulation]
        for additional options.

    """

    def __bloqade_ir__(self):
        from bloqade.ir.control.pulse import rabi

        return rabi.amplitude


class RabiPhase(Field):
    """
    This node represent phase of a rabi field.

    Examples:

        - To specify rabi phase of rydberg coupling:

        >>> node = bloqade.start.rydberg.rabi.phase
        >>> type(node)
        <class 'bloqade.builder.field.Phase'>

        - To specify rabi phase of hyperfine coupling:

        >>> node = bloqade.start.hyperfine.rabi.phase
        >>> type(node)
        <class 'bloqade.builder.field.Phase'>

    Note:
        This node is a SpatialModulation node.
        See [`SpatialModulation`][bloqade.builder.field.SpatialModulation]
        for additional options.
    """

    def __bloqade_ir__(self):
        from bloqade.ir.control.pulse import rabi

        return rabi.phase



from bloqade.builder.base import ParamType
from bloqade.submission.ir.parallel import ParallelDecoder
from bloqade.task.base import Geometry, RemoteTask
from bloqade.submission.ir.task_specification import QuEraTaskSpecification
from bloqade.submission.braket import BraketBackend

from bloqade.submission.base import ValidationError
from bloqade.submission.ir.task_results import QuEraTaskResults, QuEraTaskStatusCode
import warnings
from dataclasses import dataclass
from typing import Dict, Optional


## keep the old conversion for now,
## we will remove conversion btwn QuEraTask <-> BraketTask,
## and specialize/dispatching here.
@dataclass
class BraketTask(RemoteTask):
    task_id: Optional[str]
    backend: BraketBackend
    task_ir: QuEraTaskSpecification
    metadata: Dict[str, ParamType]
    parallel_decoder: Optional[ParallelDecoder]
    task_result_ir: Optional[QuEraTaskResults] = None

    def submit(self, force: bool = False) -> "BraketTask":
        if not force:
            if self.task_id is not None:
                raise ValueError(
                    "the task is already submitted with %s" % (self.task_id)
                )
        self.task_id = self.backend.submit_task(self.task_ir)

        self.task_result_ir = QuEraTaskResults(task_status=QuEraTaskStatusCode.Enqueued)

        return self

    def validate(self) -> str:
        try:
            self.backend.validate_task(self.task_ir)
        except ValidationError as e:
            return str(e)

        return ""

    def fetch(self) -> "BraketTask":
        # non-blocking, pull only when its completed
        if self.task_id is None:
            raise ValueError("Task ID not found.")

        status = self.status()
        if status == QuEraTaskStatusCode.Completed:
            self.task_result_ir = self.backend.task_results(self.task_id)
        else:
            self.task_result_ir = QuEraTaskResults(task_status=status)

        return self

    def pull(self) -> "BraketTask":
        # blocking, force pulling, even its completed
        if self.task_id is None:
            raise ValueError("Task ID not found.")

        self.task_result_ir = self.backend.task_results(self.task_id)

        return self

    def result(self) -> QuEraTaskResults:
        # blocking, caching

        if self.task_result_ir is None:
            pass
        else:
            if (
                self.task_id is not None
                and self.task_result_ir.task_status != QuEraTaskStatusCode.Completed
            ):
                self.pull()

        return self.task_result_ir

    def status(self) -> QuEraTaskStatusCode:
        if self.task_id is None:
            return QuEraTaskStatusCode.Unaccepted

        return self.backend.task_status(self.task_id)

    def cancel(self) -> None:
        if self.task_id is None:
            warnings.warn("Cannot cancel task, missing task id.")
            return

        self.backend.cancel_task(self.task_id)

    @property
    def nshots(self):
        return self.task_ir.nshots

    def _geometry(self) -> Geometry:
        return Geometry(
            sites=self.task_ir.lattice.sites,
            filling=self.task_ir.lattice.filling,
            parallel_decoder=self.parallel_decoder,
        )

    def _result_exists(self) -> bool:
        if self.task_id is None:
            return False

        if self.task_result_ir is None:
            return False
        else:
            if self.task_result_ir.task_status == QuEraTaskStatusCode.Completed:
                return True
            else:
                return False

    # def submit_no_task_id(self) -> "HardwareTaskShotResults":
    #    return HardwareTaskShotResults(hardware_task=self)



# from numbers import Real
from typing import TYPE_CHECKING, Union
from bokeh.io import show
from bokeh.layouts import row

if TYPE_CHECKING:
    from bloqade.ir.location.base import AtomArrangement, ParallelRegister
    from bloqade.ir import Sequence


# NOTE: this is just a dummy type bundle geometry and sequence
#       information together and forward them to backends.
class AnalogCircuit:

    """Program is a dummy type that bundle register and sequence together."""

    def __init__(
        self,
        register: Union["AtomArrangement", "ParallelRegister"],
        sequence: "Sequence",
    ):
        self._sequence = sequence
        self._register = register

    @property
    def register(self):
        """Get the register of the program.

        Returns:
            register (Union["AtomArrangement", "ParallelRegister"])

        Note:
            If the program is built with
            [`parallelize()`][bloqade.builder.emit.Emit.parallelize],
            The the register will be a
            [`ParallelRegister`][bloqade.ir.location.base.ParallelRegister].
            Otherwise it will be a
            [`AtomArrangement`][bloqade.ir.location.base.AtomArrangement].
        """
        return self._register

    @property
    def sequence(self):
        """Get the sequence of the program.


        Returns:
            Sequence: the sequence of the program.
                See also [`Sequence`][bloqade.ir.control.sequence.Sequence].

        """
        return self._sequence

    def __eq__(self, other):
        if isinstance(other, AnalogCircuit):
            return (self.register == other.register) and (
                self.sequence == other.sequence
            )

        return False

    def __repr__(self):
        # TODO: add repr for static_params, batch_params and order
        out = ""
        if self._register is not None:
            out += self._register.__repr__()

        out += "\n"

        if self._sequence is not None:
            out += self._sequence.__repr__()

        return out

    def figure(self, **assignments):
        fig_reg = self._register.figure(**assignments)
        fig_seq = self._sequence.figure(**assignments)
        return row(fig_seq, fig_reg)

    def show(self, **assignments):
        """Interactive visualization of the program

        Args:
            **assignments: assigning the instance value (literal) to the
                existing variables in the program

        """
        show(self.figure(**assignments))


# class BraketService:
#     def __init__(self, batch: Batch):
#         self.batch = batch

#     @property
#     def aquila(self) -> "BraketHardware":
#         return BraketHardware(
#             self.batch, "arn:aws:braket:us-east-1::device/qpu/quera/Aquila"
#         )

#     @property
#     def local_emulator(self) -> "BraketLocalEmulator":
#         return BraketLocalEmulator(self.batch)


# class BraketHardware:
#     def __init__(self, program: Program, device_arn: str):
#         self.program = program
#         self.device_arn = device_arn

#     def _compile_batch(
#         self, shots: int, args: Tuple[Real, ...], name: Optional[str] = None
#     ) -> "RemoteBatch":
#         from bloqade.compile.braket import BraketBatchCompiler
#         from bloqade.submission.braket import BraketBackend

#         backend = BraketBackend(
#             device_arn=self.device_arn,
#         )

#         return BraketBatchCompiler(
#             program=self.program,
#             backend=backend,
#         ).compile(shots, args, name=name)

#     def submit(
#         self, shots: int, args: Tuple[Real, ...], name: Optional[str] = None
#     ) -> "RemoteBatch":
#         batch = self._compile_batch(shots, args, name=name)
#         batch._submit()

#         return batch

#     def run(
#         self,
#         shots: int,
#         args: Tuple[Real, ...],
#         name: Optional[str] = None,
#         shuffle_submit_order: bool = False,
#     ) -> "RemoteBatch":
#         batch = self._compile_batch(shots, args, name=name)
#         batch._submit(shuffle_submit_order=shuffle_submit_order)
#         batch.pull()

#         return batch

#     def __call__(self, *args, shots: int = 1):
#         return self.run(shots, args)



from typing import Optional, Union, List
from bloqade.builder.base import Builder
from bloqade.builder.pragmas import Parallelizable, Flattenable, BatchAssignable
from bloqade.builder.backend import BackendRoute
from bloqade.builder.parse.trait import Parse
import numpy as np
from numbers import Real
from decimal import Decimal


def cast_param(value: Union[Real, List[Real]]):
    if isinstance(value, list):
        return list(map(Decimal, map(str, value)))

    if isinstance(value, (Real, Decimal)):
        return Decimal(str(value))

    raise ValueError("value must be a real number or a list of real numbers")


class AssignBase(Builder):
    __match_args__ = ("_assignments", "__parent__")

    def __init__(self, parent: Optional[Builder] = None, **assignments) -> None:
        super().__init__(parent)
        # TODO: implement checks for assignments
        self._assignments = assignments


class Assign(
    AssignBase, BatchAssignable, Flattenable, Parallelizable, BackendRoute, Parse
):
    def __init__(self, parent: Optional[Builder] = None, **assignments) -> None:
        for key, value in assignments.items():
            assignments[key] = cast_param(value)
        super().__init__(parent, **assignments)


class BatchAssign(AssignBase, Parallelizable, BackendRoute, Parse):
    def __init__(self, parent: Optional[Builder] = None, **assignments) -> None:
        if not len(np.unique(list(map(len, assignments.values())))) == 1:
            raise ValueError(
                "all the assignment variables need to have same number of elements."
            )

        for key, values in assignments.items():
            assignments[key] = list(map(cast_param, values))

        super().__init__(parent, **assignments)
